#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2020-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause
"""Contains support for BEE encryption for RT10xx devices."""


import logging
import os
from struct import calcsize, pack, unpack_from
from typing import Any, Dict, List, Optional, Sequence

from Crypto.Cipher import AES

from spsdk import SPSDKError
from spsdk.exceptions import SPSDKOverlapError
from spsdk.utils import UTILS_DATA_FOLDER
from spsdk.utils.crypto import Counter, crypto_backend
from spsdk.utils.database import Database
from spsdk.utils.easy_enum import Enum
from spsdk.utils.misc import (
    DebugInfo,
    align_block_fill_random,
    extend_block,
    load_binary,
    split_data,
    value_to_bytes,
    value_to_int,
)
from spsdk.utils.schema_validator import ConfigTemplate, ValidationSchemas

logger = logging.getLogger(__name__)


BEE_DATA_FOLDER: str = os.path.join(UTILS_DATA_FOLDER, "bee")
BEE_SCH_FILE: str = os.path.join(BEE_DATA_FOLDER, "sch_bee.yaml")
BEE_DATABASE_FILE: str = os.path.join(BEE_DATA_FOLDER, "database.yaml")

# maximal size of encrypted block in bytes
BEE_ENCR_BLOCK_SIZE = 0x400
# mask of the bits in the address, that must be zero
_ENCR_BLOCK_ADDR_MASK = BEE_ENCR_BLOCK_SIZE - 1  # 0x3FF


class BeeBaseClass:
    """BEE base class."""

    # format of the binary representation of the class, used as parameter for struct.pack/unpack methods
    # the format is class specific, and must be defined in child class
    _FORMAT = "@_must_be_defined_in_child_class_@"

    @classmethod
    def _struct_format(cls) -> str:
        """:return: format string for struct.pack/unpack function used for export/import from binary format."""
        return cls._FORMAT  # _FORMAT class constant must be defined in child class

    @classmethod
    def get_size(cls) -> int:
        """:return: size of the exported binary data in bytes."""
        return calcsize(cls._struct_format())

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, self.__class__) and (vars(other) == vars(self))

    @property
    def size(self) -> int:
        """:return: size of the exported binary data in bytes."""
        return self.get_size()

    def info(self) -> str:
        """Info method.

        :return: text description of the instance.
        :raises NotImplementedError: Derived class has to implement this method
        """
        raise NotImplementedError("Derived class has to implement this method.")

    def update(self) -> None:
        """Updates internal fields of the instance."""

    def validate(self) -> None:
        """Validates the configuration of the instance.

        It is recommended to call the method before export and after parsing.
        """

    def export(self) -> bytes:
        """:return: binary representation of the region (serialization)."""
        self.update()
        self.validate()
        return b""

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> Any:
        """Deserialization.

        :param data: binary data to be parsed
        :param offset: to start parsing the data
        :return: instance created from binary data; this method returns just `0`
        :raises SPSDKError: If size of the data is not sufficient
        """
        if len(data) - offset < cls.get_size():
            raise SPSDKError("Insufficient size of the data")
        return 0


class BeeFacRegion(BeeBaseClass):
    """BEE Factory Access Control (FAC) region."""

    # format of the binary representation of the class, used as parameter for struct.pack/unpack methods
    _FORMAT = "<3I20s"

    def __init__(self, start: int = 0, length: int = 0, protected_level: int = 0):
        """Constructor.

        :param start: Start address of one FAC region, align at 1KB boundary; 32-bit number
        :param length: Length of one FAC region, align at 1KB boundary; 32-bit number
        :param protected_level: Protected level: 0/1/2/3; 32-bit number
        """
        self.start_addr = start
        self.length = length
        self.protected_level = protected_level
        # immediately validate all parameters
        self.validate()

    def __str__(self) -> str:
        return f"FAC: 0x{self.start_addr:08x}[0x{self.length:x}]"

    @property
    def end_addr(self) -> int:
        """:return: end address of the region (which is last address of the region + 1)."""
        return self.start_addr + self.length

    def info(self) -> str:
        """:return: test description of the instance."""
        return f"FAC(start={hex(self.start_addr)}, length={hex(self.length)}, protected_level={self.protected_level})"

    def validate(self) -> None:
        """Validates the configuration of the instance."""
        if (self.start_addr & _ENCR_BLOCK_ADDR_MASK != 0) and (
            self.length & _ENCR_BLOCK_ADDR_MASK != 0
        ):
            raise SPSDKError("Invalid configuration of the instance")
        if self.protected_level < 0 or self.protected_level > 3:
            raise SPSDKError("Invalid protected level")
        if self.start_addr < 0 or self.end_addr > 0xFFFFFFFF or self.start_addr >= self.end_addr:
            raise SPSDKError("Invalid start/end address")

    def export(self) -> bytes:
        """Exports the binary representation."""
        result = super().export()
        return result + pack(
            self._struct_format(),
            self.start_addr,
            self.end_addr,
            self.protected_level,
            b"\x00" * 20,
        )

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> "BeeFacRegion":
        """Deserialization.

        :param data: binary data to be parsed
        :param offset: to start parsing the data
        :return: instance created from binary data
        :raises SPSDKError: If reserved area is non-zero
        """
        super().parse(data, offset)  # check size of the data
        (start, end, protected_level, _reserved) = unpack_from(
            BeeFacRegion._struct_format(), data, offset
        )
        if _reserved != b"\x00" * 20:
            raise SPSDKError("Reserved area is non-zero")
        return BeeFacRegion(start, end - start, protected_level)


class BeeProtectRegionBlockAesMode(Enum):
    """AES mode selection for BEE PRDB encryption."""

    ECB = 0
    CTR = 1


class BeeProtectRegionBlock(BeeBaseClass):
    """BEE protect region block (PRDB)."""

    # format of the binary representation of the class, used as parameter for struct.pack/unpack methods
    _FORMAT = "<8I16s32s"
    # low TAG used in the header
    TAGL = 0x5F474154  # "TAG_"
    # high TAG used in the header
    TAGH = 0x52444845  # "EHDR"
    # version of the format
    VERSION = 0x56010000
    # number of FAC regions included in the header; the real limit of FAC regions depends on the processor
    FAC_REGIONS = 4
    # total size
    SIZE = 0x100

    @classmethod
    def get_size(cls) -> int:
        """:return: size of the exported binary data in bytes."""
        return cls.SIZE

    def __init__(
        self,
        encr_mode: BeeProtectRegionBlockAesMode = BeeProtectRegionBlockAesMode.CTR,
        lock_options: int = 0,
        counter: Optional[bytes] = None,
    ):
        """Constructor.

        :param encr_mode: AES encryption mode
        :param lock_options: Lock options; 32-bit number
        :param counter: Counter for AES-CTR mode; 16 bytes; by default, random value is used
        """
        # - Encrypt region info:
        self._start_addr = 0  # this is calculated automatically based on FAC regions
        self._end_addr = 0xFFFFFFFF  # this is calculated automatically based on FAC regions
        self.mode = encr_mode
        self.lock_options = lock_options
        self.counter = (
            counter if counter else crypto_backend().random_bytes(12) + b"\x00\x00\x00\x00"
        )

        # - FAC regions, 1 - 4
        self.fac_regions: List[BeeFacRegion] = []

    def update(self) -> None:
        """Updates start and end address of the encryption region."""
        super().update()
        # update FAC regions
        for fac in self.fac_regions:
            fac.update()
        # update start and end address
        min_addr = 0 if self.fac_count == 0 else 0xFFFFFFFF
        max_addr = 0
        for fac in self.fac_regions:
            min_addr = min(min_addr, fac.start_addr)
            max_addr = max(max_addr, fac.end_addr)
        self._start_addr = min_addr
        self._end_addr = max_addr

    def add_fac(self, fac: BeeFacRegion) -> None:
        """Append FAC region.

        :param fac: Factory Access Control to be added
        """
        self.fac_regions.append(fac)
        self.update()

    @property
    def fac_count(self) -> int:
        """:return: number of Factory Access Control regions."""
        return len(self.fac_regions)

    def info(self) -> str:
        """:return: test description of the instance."""
        result = f"BEE Region Header (start={hex(self._start_addr)}, end={hex(self._end_addr)})\n"
        result += f"AES Encryption mode: {BeeProtectRegionBlockAesMode.name(self.mode)}\n"
        for fac in self.fac_regions:
            result += fac.info() + "\n"
        return result

    def validate(self) -> None:
        """Validates settings of the instance."""
        if self._start_addr < 0 or self._start_addr > 0xFFFFFFFF:
            raise SPSDKError("Invalid start address")
        if self._start_addr > self._end_addr or self._end_addr > 0xFFFFFFFF:
            raise SPSDKError("Invalid start/end address")
        if self.mode != BeeProtectRegionBlockAesMode.CTR:
            raise SPSDKError("Only AES/CTR encryption mode supported now")
        if len(self.counter) != 16:
            raise SPSDKError("Invalid counter")
        if self.counter[-4:] != b"\x00\x00\x00\x00":
            raise SPSDKError("last four bytes must be zero")
        if self.fac_count <= 0 or self.fac_count > self.FAC_REGIONS:
            raise SPSDKError("Invalid FAC regions")
        for fac in self.fac_regions:
            fac.validate()

    def export(self) -> bytes:
        """:return: binary representation of the region (serialization)."""
        result = super().export()
        result += pack(
            self._struct_format(),
            self.TAGL,
            self.TAGH,
            self.VERSION,
            self.fac_count,
            self._start_addr,
            self._end_addr,
            self.mode,
            self.lock_options,
            self.counter[::-1],  # bytes swapped: reversed order
            b"\x00" * 32,
        )
        for fac in self.fac_regions:
            result += fac.export()
        result = extend_block(result, self.SIZE)
        return result

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> "BeeProtectRegionBlock":
        """Deserialization.

        :param data: binary data to be parsed
        :param offset: to start parsing the data
        :return: instance created from binary data
        :raises SPSDKError: If format does not match
        """
        super().parse(data, offset)  # check size of the input data
        (
            tagl,
            tagh,
            version,
            fac_count,
            start_addr,
            end_addr,
            mode,
            lock_options,
            counter,
            _reserved_32,
        ) = unpack_from(BeeProtectRegionBlock._struct_format(), data, offset)
        #
        if (tagl != BeeProtectRegionBlock.TAGL) or (tagh != BeeProtectRegionBlock.TAGH):
            raise SPSDKError("Invalid tag or unsupported version")
        if version != BeeProtectRegionBlock.VERSION:
            raise SPSDKError("Unsupported version")
        if _reserved_32 != b"\x00" * 32:
            raise SPSDKError("Reserved area is non-zero")
        #
        result = BeeProtectRegionBlock(mode, lock_options, counter[::-1])
        result._start_addr = start_addr
        result._end_addr = end_addr
        offset += calcsize(BeeProtectRegionBlock._struct_format())
        for _ in range(fac_count):
            fac = BeeFacRegion.parse(data, offset)
            result.add_fac(fac)
            offset += fac.size
        result.validate()
        return result

    def is_inside_region(self, start_addr: int) -> bool:
        """Returns true if the start address lies within any FAC region.

        :param start_addr: start address of the data
        """
        return self._start_addr <= start_addr < self._end_addr

    def encrypt_block(self, key: bytes, start_addr: int, data: bytes) -> bytes:
        """Encrypt block located in any FAC region.

        :param key: user for encryption
        :param start_addr: start address of the data
        :param data: binary block to be encrypted; the block size must be BEE_ENCR_BLOCK_SIZE
        :return: encrypted block if it is inside any FAC region; untouched block if it is not in any FAC region
        :raises SPSDKError: When incorrect length of binary block
        :raises SPSDKError: When encryption mode different from AES/CTR provided
        :raises SPSDKError: When invalid length of key
        :raises SPSDKError: When invalid range of region
        """
        if len(data) > BEE_ENCR_BLOCK_SIZE:
            raise SPSDKError("Incorrect length of binary block to be encrypted")
        if self.is_inside_region(start_addr):
            if self.mode != BeeProtectRegionBlockAesMode.CTR:
                raise SPSDKError("only AES/CTR encryption mode supported now")
            if len(key) != 16:
                raise SPSDKError("Invalid length of key")
            for fac in self.fac_regions:
                if fac.start_addr <= start_addr < fac.end_addr:
                    if start_addr + len(data) > fac.end_addr:
                        raise SPSDKError("Invalid range of region")
                    cntr_key = Counter(
                        self.counter,
                        ctr_value=start_addr >> 4,
                        ctr_byteorder_encoding="big",
                    )
                    logger.debug(
                        f"Encrypting data, start={hex(start_addr)},"
                        f"end={hex(start_addr + len(data))} with {self.info()} using fac {fac.info()}"
                    )
                    data = align_block_fill_random(data, 16)  # align data to 16 bytes
                    return crypto_backend().aes_ctr_encrypt(key, data, cntr_key.value)
        return data


class BeeKIB(BeeBaseClass):
    """BEE Key block.

    Contains keys used to encrypt PRDB content.
    """

    # key length in bytes
    _KEY_LEN = 16
    # format of the binary representation of the class, used as parameter for struct.pack/unpack methods
    _FORMAT = "16s16s"

    def __init__(self, kib_key: Optional[bytes] = None, kib_iv: Optional[bytes] = None):
        """Constructor.

        :param kib_key: AES key
        :param kib_iv: AES initialization vector
        """
        # Key Info Block (KIB)
        self.kib_key = kib_key if kib_key else crypto_backend().random_bytes(16)
        self.kib_iv = kib_iv if kib_iv else crypto_backend().random_bytes(16)

    def info(self) -> str:
        """:return: test description of the instance."""
        return f"BEE-KIB: {self.kib_key.hex()}, {self.kib_iv.hex()}"

    def validate(self) -> None:
        """Validates settings of the instance.

        :raises SPSDKError: If invalid length of kib key
        :raises SPSDKError: If invalid length of kib iv
        """
        if len(self.kib_key) != self._KEY_LEN:
            raise SPSDKError("Invalid length of kib key")
        if len(self.kib_iv) != self._KEY_LEN:
            raise SPSDKError("Invalid length of kib iv")

    def export(self) -> bytes:
        """Exports binary representation of the region (serialization)."""
        result = super().export()
        return result + pack(self._struct_format(), self.kib_key, self.kib_iv)

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> "BeeKIB":
        """Deserialization.

        :param data: binary data to be parsed
        :param offset: to start parsing the data
        :return: instance created from binary data
        """
        super().parse(data, offset)  # check size of the input data
        (key, iv) = unpack_from(BeeKIB._struct_format(), data, offset)
        result = cls(key, iv)
        result.validate()
        return result


class BeeRegionHeader(BeeBaseClass):
    """BEE keys and regions header."""

    # offset of the Protected Region Block in the header
    PRDB_OFFSET = 0x80
    # total size including padding
    SIZE = 0x200

    @classmethod
    def _struct_format(cls) -> str:
        """:raises SPSDKError: It is not expected to called for the class."""
        raise SPSDKError(
            "This method is not expected to be used for this class, format depends on its fields"
        )

    @classmethod
    def get_size(cls) -> int:
        """:return: size of the exported binary data in bytes."""
        return cls.SIZE

    def __init__(
        self,
        prdb: Optional[BeeProtectRegionBlock] = None,
        sw_key: Optional[bytes] = None,
        kib: Optional[BeeKIB] = None,
    ):
        """Constructor.

        :param prdb: protect region block; None to use default
        :param sw_key: key used to encrypt KIB content
        :param kib: keys block; None to use default
        """
        self._prdb = prdb if (prdb is not None) else BeeProtectRegionBlock()
        self._sw_key = sw_key if (sw_key is not None) else crypto_backend().random_bytes(16)
        self._kib = kib if (kib is not None) else BeeKIB()

    def add_fac(self, fac: BeeFacRegion) -> None:
        """Append FAC region.

        :param fac: to be added
        """
        self._prdb.add_fac(fac)

    @property
    def fac_regions(self) -> Sequence[BeeFacRegion]:
        """:return: lift of Factory Access Control regions."""
        return self._prdb.fac_regions

    def info(self) -> str:
        """:return: test description of the instance."""
        result = "BEE Region Header\n"
        result += f"- KIB: {self._kib.info()}"
        result += f"- PRDB: {self._prdb.info()}"
        return result

    def sw_key_fuses(self) -> Sequence[int]:
        """:return: sequence of fuse values for SW key to be burned into processor.

        The result is ordered, first value should be burned to the lowest address.
        """
        result = []
        for pos in range(16, 0, -4):
            result.append(unpack_from(">I", self._sw_key[pos - 4 : pos])[0])
        return result

    def update(self) -> None:
        """Updates internal fields of the instance."""
        super().update()
        self._kib.update()
        self._prdb.update()

    def validate(self) -> None:
        """Validates settings of the instance.

        :raises SPSDKError: If settings invalid
        """
        self._kib.validate()
        self._prdb.validate()
        if len(self._sw_key) != 16:
            raise SPSDKError("Invalid settings")

    def export(self, dbg_info: DebugInfo = DebugInfo.disabled()) -> bytes:
        """Serialization to binary representation.

        :param dbg_info: instance allowing to provide debug info about exported data
        :return: binary representation of the region (serialization).
        """
        result = super().export()
        # KIB
        kib_data = self._kib.export()
        dbg_info.append_binary_section("BEE-KIB (non-crypted)", kib_data)
        aes_ecb = AES.new(self._sw_key, AES.MODE_ECB)
        result += aes_ecb.encrypt(kib_data)
        # padding
        result = extend_block(result, self.PRDB_OFFSET)
        # PRDB
        prdb_data = self._prdb.export()
        dbg_info.append_binary_section("BEE-PRDB (non-crypted)", prdb_data)
        aes_cbc = AES.new(self._kib.kib_key, AES.MODE_CBC, self._kib.kib_iv)
        result += aes_cbc.encrypt(prdb_data)
        # padding
        return extend_block(result, self.SIZE)

    @classmethod
    def parse(cls, data: bytes, offset: int = 0, sw_key: bytes = b"") -> "BeeRegionHeader":
        """Deserialization.

        :param data: binary data to be parsed
        :param offset: to start parsing the data
        :param sw_key: SW key used to decrypt the EKIB data (the key is marked as SW_GP2 on RT10xx)
        :return: instance created from binary data
        :raises SPSDKError: If invalid sw key
        """
        super().parse(data, offset)  # check size of the input data
        if len(sw_key) != 16:
            raise SPSDKError("Invalid sw key")
        aes_ecb = AES.new(sw_key, AES.MODE_ECB)
        decr_data = aes_ecb.decrypt(data[offset : offset + BeeKIB.get_size()])
        kib = BeeKIB.parse(decr_data)
        aes_cbc = AES.new(kib.kib_key, AES.MODE_CBC, kib.kib_iv)
        decr_data = aes_cbc.decrypt(
            data[offset + cls.PRDB_OFFSET : offset + cls.PRDB_OFFSET + BeeProtectRegionBlock.SIZE]
        )
        prdb = BeeProtectRegionBlock.parse(decr_data)
        result = cls(prdb, sw_key, kib)
        result.validate()
        return result

    def is_inside_region(self, start_addr: int) -> bool:
        """Returns true if the start address lies within any FAC region.

        :param start_addr: start address of the data
        """
        return self._prdb.is_inside_region(start_addr)

    def encrypt_block(self, start_addr: int, data: bytes) -> bytes:
        """Encrypt block located in any FAC region.

        :param start_addr: start address of the data
        :param data: binary block to be encrypted; the block size must be BEE_ENCR_BLOCK_SIZE
        :return: encrypted block if it is inside any FAC region; untouched block if it is not in any FAC region
        """
        return self._prdb.encrypt_block(self._sw_key, start_addr, data)


class BeeNxp:
    """BeeNxp class."""

    def __init__(
        self,
        headers: List[BeeRegionHeader],
        input_image: bytes,
        base_address: int,
    ):
        """Constructor.

        :param headers: list of BEE Region Headers
        :param input_image: Input image to be encrypted
        :param base_address: Base address of the image
        """
        self.headers = headers
        self.input_image = input_image
        self.base_address = base_address

    def export_image(self) -> bytes:
        """Export encrypted binary image.

        :return: encrypted image
        """
        encrypted_data = bytearray()
        image_data = bytearray(self.input_image)
        base_address = self.base_address

        for block in split_data(image_data, BEE_ENCR_BLOCK_SIZE):
            logger.debug(f"Reading {hex(base_address)}, size={hex(len(block))}")
            for header in self.headers:
                block = header.encrypt_block(base_address, block)
            encrypted_data.extend(block)
            base_address += len(block)

        return bytes(encrypted_data)

    def export_headers(self) -> List[bytes]:
        """Export BEE headers.

        :return: BEE region headers
        """
        headers = []
        for header in self.headers:
            headers.append(header.export())

        return headers

    @staticmethod
    def get_supported_families() -> List[str]:
        """Get all supported families for BEE.

        :return: List of supported families.
        """
        database = Database(BEE_DATABASE_FILE)
        return database.devices.device_names

    @staticmethod
    def get_validation_schemas() -> List[Dict[str, Any]]:
        """Get list of validation schemas.

        :return: Validation list of schemas.
        """
        schemas = ValidationSchemas.get_schema_file(BEE_SCH_FILE)
        return [schemas["bee_output"], schemas["bee"]]

    @staticmethod
    def generate_config_template() -> str:
        """Generate BEE configuration template.

        :return: Dictionary of individual templates (key is name of template, value is template itself).
        """
        val_schemas = BeeNxp.get_validation_schemas()

        yaml_data = ConfigTemplate(
            "BEE configuration template",
            val_schemas,
        ).export_to_yaml()

        return yaml_data

    @staticmethod
    def check_overlaps(bee_headers: List[BeeRegionHeader], start_addr: int) -> None:
        """Check for overlaps in regions.

        :param bee_headers: List of BeeRegionHeader
        :param start_addr: start address of a region to be checked
        :raises SPSDKOverlapError: if the address is inside any region
        """
        for header in bee_headers:
            for region in header.fac_regions:
                if region.start_addr <= start_addr < region.end_addr:
                    raise SPSDKOverlapError(
                        f"Region start address {hex(start_addr)} is overlapping with {region.info()}"
                    )

    @staticmethod
    def load_from_config(
        config: Dict[str, Any], search_paths: Optional[List[str]] = None
    ) -> "BeeNxp":
        """Converts the configuration into an BEE image object.

        "config" contains dictionary of configurations.

        :raises SPSDKError: if the count of BEE engines is invalid.
        :param config: Configuration dictionary.
        :param search_paths: List of paths where to search for the file, defaults to None
        :return: initialized BeeNxp object.
        """
        input_binary = load_binary(config["input_binary"], search_paths)

        engine_selection = config["engine_selection"]
        bee_engines: List[Dict[str, Any]] = config["bee_engine"]
        base_address = value_to_int(config["base_address"])

        prdb = BeeProtectRegionBlock()
        kib = BeeKIB()

        bee_headers: List[BeeRegionHeader] = []

        engine_selections = {"engine0": [0], "engine1": [1], "both": [0, 1]}

        try:
            for engine_idx in engine_selections[engine_selection]:
                key = value_to_bytes(bee_engines[engine_idx]["user_key"], byte_cnt=16)
                if bee_engines[engine_idx].get("header_path"):
                    bin_ehdr = load_binary(bee_engines[engine_idx]["header_path"], search_paths)
                    ehdr = BeeRegionHeader.parse(bin_ehdr, sw_key=key)
                else:
                    ehdr = BeeRegionHeader(prdb, key, kib)
                bee_headers.append(ehdr)
        except IndexError as exc:
            raise SPSDKError("The count of BEE engines is invalid") from exc

        for engine_idx, bee_engine in enumerate(bee_engines):
            for protected_region in bee_engine.get("protected_region", []):
                fac = BeeFacRegion(
                    value_to_int(protected_region["start_address"]),
                    value_to_int(protected_region["length"]),
                    value_to_int(protected_region["protected_level"]),
                )
                BeeNxp.check_overlaps(bee_headers, fac.start_addr)
                bee_headers[engine_idx].add_fac(fac)

        bee = BeeNxp(bee_headers, input_binary, base_address)

        return bee



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2022-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause
"""Module to keep additional utilities for binary images."""

import logging
import math
import os
import textwrap
from typing import TYPE_CHECKING, Any, Dict, List, Optional

import colorama

from spsdk import SPSDK_DATA_FOLDER
from spsdk.exceptions import SPSDKError, SPSDKOverlapError, SPSDKValueError
from spsdk.utils.misc import (
    BinaryPattern,
    align,
    align_block,
    find_file,
    format_value,
    size_fmt,
    write_file,
)
from spsdk.utils.schema_validator import ConfigTemplate, ValidationSchemas

if TYPE_CHECKING:
    # bincopy will be loaded lazily as needed, this is just to satisfy type-hint checkers
    import bincopy

BINARY_SCH_FILE = os.path.join(SPSDK_DATA_FOLDER, "image", "sch_binary.yaml")

logger = logging.getLogger(__name__)


class ColorPicker:
    """Simple class to get each time when ask different color from list."""

    COLORS = [
        colorama.Fore.LIGHTBLACK_EX,
        colorama.Fore.BLUE,
        colorama.Fore.GREEN,
        colorama.Fore.CYAN,
        colorama.Fore.YELLOW,
        colorama.Fore.MAGENTA,
        colorama.Fore.WHITE,
        colorama.Fore.LIGHTBLUE_EX,
        colorama.Fore.LIGHTCYAN_EX,
        colorama.Fore.LIGHTGREEN_EX,
        colorama.Fore.LIGHTMAGENTA_EX,
        colorama.Fore.LIGHTWHITE_EX,
        colorama.Fore.LIGHTYELLOW_EX,
    ]

    def __init__(self) -> None:
        """Constructor of ColorPicker."""
        self.index = len(self.COLORS)

    def get_color(self, unwanted_color: Optional[str] = None) -> str:
        """Get new color from list.

        :param unwanted_color: Color that should be omitted.
        :return: Color
        """
        self.index += 1
        if self.index >= len(ColorPicker.COLORS):
            self.index = 0
        if unwanted_color and ColorPicker.COLORS[self.index] == unwanted_color:
            return self.get_color(unwanted_color)
        return ColorPicker.COLORS[self.index]


class BinaryImage:
    """Binary Image class."""

    MINIMAL_DRAW_WIDTH = 30

    def __init__(
        self,
        name: str,
        size: int = 0,
        offset: int = 0,
        description: Optional[str] = None,
        binary: Optional[bytes] = None,
        pattern: Optional[BinaryPattern] = None,
        alignment: int = 1,
        parent: Optional["BinaryImage"] = None,
    ) -> None:
        """Binary Image class constructor.

        :param name: Name of Image.
        :param size: Image size.
        :param offset: Image offset in parent image, defaults to 0
        :param description: Text description of image, defaults to None
        :param binary: Optional binary content.
        :param pattern: Optional binary pattern.
        :param alignment: Optional alignment of result image
        :param parent: Handle to parent object, defaults to None
        """
        self.name = name
        self.description = description
        self.offset = offset
        self._size = align(size, alignment)
        self.binary = binary
        self.pattern = pattern
        self.alignment = alignment
        self.parent = parent

        if parent:
            assert isinstance(parent, BinaryImage)
        self.sub_images: List["BinaryImage"] = []

    @property
    def size(self) -> int:
        """Size property."""
        return self._size

    @size.setter
    def size(self, value: int) -> None:
        """Size property setter."""
        self._size = align(value, self.alignment)

    def add_image(self, image: "BinaryImage") -> None:
        """Add new sub image information.

        :param image: Image object.
        """
        image.parent = self
        for i, child in enumerate(self.sub_images):
            if image.offset < child.offset:
                self.sub_images.insert(i, image)
                return
        self.sub_images.append(image)

    def join_images(self) -> None:
        """Join all sub images into main binary block."""
        binary = self.export()
        self.sub_images.clear()
        self.binary = binary

    @property
    def image_name(self) -> str:
        """Image name including all parents.

        :return: Full Image name
        """
        if self.parent:
            return self.parent.image_name + "=>" + self.name
        return self.name

    @property
    def absolute_address(self) -> int:
        """Image absolute address relative to base parent.

        :return: Absolute address relative to base parent
        """
        if self.parent:
            return self.parent.absolute_address + self.offset
        return self.offset

    def aligned_start(self, alignment: int = 4) -> int:
        """Returns aligned start address.

        :param alignment: The alignment value, defaults to 4.
        :return: Floor alignment address.
        """
        return math.floor(self.absolute_address / alignment) * alignment

    def aligned_length(self, alignment: int = 4) -> int:
        """Returns aligned length for erasing purposes.

        :param alignment: The alignment value, defaults to 4.
        :return: Ceil alignment length.
        """
        end_address = self.absolute_address + len(self)
        aligned_end = math.ceil(end_address / alignment) * alignment
        aligned_len = aligned_end - self.aligned_start(alignment)
        return aligned_len

    def info(self) -> str:
        """Provides information about image.

        :return: String information about Image.
        """
        size = len(self)
        ret = ""
        ret += f"Name:      {self.image_name}\n"
        ret += f"Starts:    {hex(self.absolute_address)}\n"
        ret += f"Ends:      {hex(self.absolute_address+ size-1)}\n"
        ret += f"Size:      {size_fmt(size, use_kibibyte=False)}\n"
        ret += f"Alignment: {size_fmt(self.alignment, use_kibibyte=False)}\n"
        if self.pattern:
            ret += f"Pattern:{self.pattern.pattern}\n"
        if self.description:
            ret += self.description + "\n"
        return ret

    def validate(self) -> None:
        """Validate if the images doesn't overlaps each other."""
        if self.offset < 0:
            raise SPSDKValueError(
                f"Image offset of {self.image_name} cannot be in negative numbers."
            )
        if len(self) <= 0:
            raise SPSDKValueError(
                f"Image size of {self.image_name} cannot be in negative numbers or zero."
            )
        for image in self.sub_images:
            image.validate()
            begin = image.offset
            end = begin + len(image) - 1
            # Check if it fits inside the parent image
            if end >= len(self):
                raise SPSDKOverlapError(
                    f"The image {image.name} doesn't fit into {self.name} parent image."
                )
            # Check if it doesn't overlap any other sibling image
            for sibling in self.sub_images:
                if sibling != image:
                    sibling_begin = sibling.offset
                    sibling_end = sibling_begin + len(sibling) - 1
                    if end < sibling_begin or begin > sibling_end:
                        continue

                    raise SPSDKOverlapError(
                        f"The image overlap error:\n"
                        f"{image.info()}\n"
                        "overlaps the:\n"
                        f"{sibling.info()}\n"
                    )

    def get_min_draw_width(self, include_sub_images: bool = True) -> int:
        """Get minimal width of table for draw function.

        :param include_sub_images: Include also sub images into, defaults to True
        :return: Minimal width in characters.
        """
        widths = [
            self.MINIMAL_DRAW_WIDTH,
            len(f"+==-0x0000_0000= {self.name} =+"),
            len(f"|Size: {size_fmt(len(self), False)}|"),
        ]
        if include_sub_images:
            for child in self.sub_images:
                widths.append(child.get_min_draw_width() + 2)  # +2 means add vertical borders
        return max(widths)

    def draw(
        self,
        include_sub_images: bool = True,
        width: int = 0,
        color: str = "",
        no_color: bool = False,
    ) -> str:
        # fmt: off
        """Draw the image into the ASCII graphics.

        :param include_sub_images: Include also sub images into, defaults to True
        :param width: Fixed width of table, 0 means autosize.
        :param color: Color of this block, None means automatic color.
        :param no_color: Disable adding colors into output.
        :raises SPSDKValueError: In case of invalid width.
        :return: ASCII art representation of image.
        """
        # +==0x0000_0000==Title1===============+
        # |            Size: 2048B             |
        # |           Description1             |
        # |       Description1 2nd line        |
        # |+==0x0000_0000==Title11============+|
        # ||           Size: 512B             ||
        # ||           Description11          ||
        # ||       Description11 2nd line     ||
        # |+==0x0000_01FF=====================+|
        # |                                    |
        # |+==0x0000_0210==Title12============+|
        # ||           Size: 512B             ||
        # ||           Description12          ||
        # ||       Description12 2nd line     ||
        # |+==0x0000_041F=====================+|
        # +==0x0000_07FF=======================+
        # fmt: on
        def _get_centered_line(text: str) -> str:
            text_len = len(text)
            spaces = width - text_len - 2
            assert spaces >= 0, "Binary Image Draw: Center line is longer than width"
            padding_l = int(spaces / 2)
            padding_r = int(spaces - padding_l)
            return color + f"|{' '*padding_l}{text}{' '*padding_r}|\n"

        def wrap_block(inner: str) -> str:
            wrapped_block = ""
            lines = inner.splitlines(keepends=False)
            for line in lines:
                wrapped_block += color + "|" + line + color + "|\n"
            return wrapped_block

        if no_color:
            color = ""
        else:
            color_picker = ColorPicker()
            try:
                self.validate()
                color = color or color_picker.get_color()
            except SPSDKError:
                color = colorama.Fore.RED

        block = "" if self.parent else "\n"
        min_width = self.get_min_draw_width(include_sub_images)
        if not width and self.parent is None:
            width = min_width

        if width < min_width:
            raise SPSDKValueError(
                f"Binary Image Draw: Width is to short ({width} < minimal width: {min_width})"
            )

        # - Title line
        header = f"+=={format_value(self.absolute_address, 32)}= {self.name} ="
        block += color + f"{header}{'='*(width-len(header)-1)}+\n"
        # - Size
        block += _get_centered_line(f"Size: {size_fmt(len(self), False)}")
        # - Description
        if self.description:
            for line in textwrap.wrap(self.description, width=width - 2, fix_sentence_endings=True):
                block += _get_centered_line(line)
        # - Pattern
        if self.pattern:
            block += _get_centered_line(f"Pattern: {self.pattern.pattern}")
        # - Inner blocks
        if include_sub_images:
            next_free_space = 0
            for child in self.sub_images:
                # If the images doesn't comes one by one place empty line
                if child.offset != next_free_space:
                    block += _get_centered_line(
                        f"Gap: {size_fmt(child.offset-next_free_space, False)}"
                    )
                next_free_space = child.offset + len(child)
                inner_block = child.draw(
                    include_sub_images=include_sub_images,
                    width=width - 2,
                    color="" if no_color else color_picker.get_color(color),
                    no_color=no_color,
                )
                block += wrap_block(inner_block)

        # - Closing line
        footer = f"+=={format_value(self.absolute_address + len(self) - 1, 32)}=="
        block += color + f"{footer}{'='*(width-len(footer)-1)}+\n"

        if self.parent is None:
            block += "\n" + "" if no_color else colorama.Fore.RESET
        return block

    def update_offsets(self) -> None:
        """Update offsets from the sub images into main offset value begin offsets."""
        offsets = []
        for image in self.sub_images:
            offsets.append(image.offset)

        min_offset = min(offsets)
        for image in self.sub_images:
            image.offset -= min_offset
        self.offset += min_offset

    def __len__(self) -> int:
        """Get length of image.

        If internal member size is not set(is zero) the size is computed from sub images.
        :return: Size of image.
        """
        if self._size:
            return self._size
        max_size = len(self.binary) if self.binary else 0
        for image in self.sub_images:
            size = image.offset + len(image)
            max_size = max(size, max_size)
        return align(max_size, self.alignment)

    def export(self) -> bytes:
        """Export represented binary image.

        :return: Byte array of binary image.
        """
        ret = bytearray(self.pattern.get_block(len(self))) if self.pattern else bytearray(len(self))
        if self.binary:
            ret[: len(self.binary)] = self.binary
        for image in self.sub_images:
            ret[image.offset : image.offset + len(image)] = image.export()[: len(image)]
        return align_block(ret, self.alignment, self.pattern)

    @staticmethod
    def get_validation_schemas() -> List[Dict[str, Any]]:
        """Get validation schemas list to check a supported configuration.

        :return: Validation schemas.
        """
        return [ValidationSchemas.get_schema_file(BINARY_SCH_FILE)]

    @staticmethod
    def load_from_config(
        config: Dict[str, Any], search_paths: Optional[List[str]] = None
    ) -> "BinaryImage":
        """Converts the configuration option into an Binary Image object.

        :param config: Description of binary image.
        :param search_paths: List of paths where to search for the file, defaults to None
        :return: Initialized Binary Image.
        """
        name = config.get("name", "Base Image")
        size = config.get("size", 0)
        pattern = BinaryPattern(config["pattern"])
        ret = BinaryImage(name=name, size=size, pattern=pattern)
        regions = config.get("regions")
        if regions:
            for i, region in enumerate(regions):
                binary_file: Dict = region.get("binary_file")
                if binary_file:
                    offset = binary_file["offset"]
                    name = binary_file.get("name", binary_file["path"])
                    ret.add_image(
                        BinaryImage.load_binary_image(
                            binary_file["path"],
                            name=name,
                            offset=offset,
                            pattern=pattern,
                            search_paths=search_paths,
                        )
                    )
                binary_block: Dict = region.get("binary_block")
                if binary_block:
                    size = binary_block["size"]
                    offset = binary_block["offset"]
                    name = binary_block.get("name", f"Binary block(#{i})")
                    pattern = BinaryPattern(binary_block["pattern"])
                    ret.add_image(BinaryImage(name, size, offset, pattern=pattern))
        return ret

    def save_binary_image(
        self,
        path: str,
        file_format: str = "BIN",
    ) -> None:
        # pylint: disable=missing-param-doc
        """Save binary data file.

        :param path: Path to the file.
        :param file_format: Format of saved file ('BIN', 'HEX', 'S19'), defaults to 'BIN'.
        :raises SPSDKValueError: The file format is invalid.
        """
        file_format = file_format.upper()
        if file_format.upper() not in ("BIN", "HEX", "S19"):
            raise SPSDKValueError(f"Invalid input file format: {file_format}")

        if file_format == "BIN":
            write_file(self.export(), path, mode="wb")
            return

        def add_into_binary(bin_image: BinaryImage) -> None:
            if bin_image.pattern:
                bin_file.add_binary(
                    bin_image.pattern.get_block(len(bin_image)),
                    address=bin_image.absolute_address,
                    overwrite=True,
                )

            if bin_image.binary:
                bin_file.add_binary(
                    bin_image.binary, address=bin_image.absolute_address, overwrite=True
                )

            for sub_image in bin_image.sub_images:
                add_into_binary(sub_image)

        # import bincopy only if needed to save startup time
        import bincopy  # pylint: disable=import-outside-toplevel

        bin_file = bincopy.BinFile()
        add_into_binary(self)

        if file_format == "HEX":
            write_file(bin_file.as_ihex(), path)
            return

        # And final supported format is....... Yes, S record from MOTOROLA
        write_file(bin_file.as_srec(), path)

    @staticmethod
    def generate_config_template() -> str:
        """Generate configuration template.

        :return: Template to create binary merge..
        """
        return ConfigTemplate(
            "Binary Image Configuration template.",
            BinaryImage.get_validation_schemas(),
        ).export_to_yaml()

    @staticmethod
    def load_binary_image(
        path: str,
        name: Optional[str] = None,
        size: int = 0,
        offset: int = 0,
        description: Optional[str] = None,
        pattern: Optional[BinaryPattern] = None,
        search_paths: Optional[List[str]] = None,
        alignment: int = 1,
    ) -> "BinaryImage":
        # pylint: disable=missing-param-doc
        r"""Load binary data file.

        :param path: Path to the file.
        :param name: Name of Image, defaults to file name.
        :param size: Image size, defaults to 0.
        :param offset: Image offset in parent image, defaults to 0
        :param description: Text description of image, defaults to None
        :param pattern: Optional binary pattern.
        :param search_paths: List of paths where to search for the file, defaults to None
        :param alignment: Optional alignment of result image
        :raises SPSDKError: The binary file cannot be loaded.
        :return: Binary data represented in BinaryImage class.
        """
        path = find_file(path, search_paths=search_paths)
        try:
            with open(path, "rb") as f:
                data = f.read(4)
        except Exception as e:
            raise SPSDKError(f"Error loading file: {str(e)}") from e

        # import bincopy only if needed to save startup time
        import bincopy  # pylint: disable=import-outside-toplevel

        bin_file = bincopy.BinFile()
        try:
            if data == b"\x7fELF":
                logger.warning("Elf file support is experimental. Take that with care.")
                bin_file.add_elf_file(path)
            else:
                try:
                    bin_file.add_file(path)
                except (UnicodeDecodeError, bincopy.UnsupportedFileFormatError):
                    bin_file.add_binary_file(path)
        except Exception as e:
            raise SPSDKError(f"Error loading file: {str(e)}") from e

        img_name = name or os.path.basename(path)
        img_size = size or 0
        img_descr = description or f"The image loaded from: {path} ."
        bin_image = BinaryImage(
            name=img_name,
            size=img_size,
            offset=offset,
            description=img_descr,
            pattern=pattern,
            alignment=alignment,
        )
        if len(bin_file.segments) == 0:
            raise SPSDKError(f"Load of {path} failed, can't be decoded.")

        for i, segment in enumerate(bin_file.segments):
            bin_image.add_image(
                BinaryImage(
                    name=f"Segment {i}",
                    size=len(segment.data),
                    offset=segment.address,
                    pattern=pattern,
                    binary=segment.data,
                    parent=bin_image,
                    alignment=alignment,
                )
            )
        # Optimize offsets in image
        bin_image.update_offsets()
        return bin_image



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2020-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Module with DebugCredential class."""

import math
from struct import calcsize, pack, unpack_from
from typing import Any, List, Optional, Type

from spsdk import SPSDKError, crypto
from spsdk.crypto.loaders import extract_public_key
from spsdk.crypto.signature_provider import SignatureProvider, get_signature_provider
from spsdk.dat.utils import ecc_key_to_bytes, ecc_public_numbers_to_bytes, rsa_key_to_bytes
from spsdk.exceptions import SPSDKValueError
from spsdk.image.ahab.ahab_container import SRKRecord, SRKTable
from spsdk.utils.crypto.backend_internal import internal_backend
from spsdk.utils.misc import find_file


class DebugCredential:
    """Base class for DebugCredential."""

    # Subclasses override the following invalid class member values
    FORMAT = "INVALID_FORMAT"
    FORMAT_NO_SIG = "INVALID_FORMAT"
    SOCC_FORMAT = ""
    VERSION = "0.0"
    HASH_LENGTH = 32

    SOCC_LIST = {
        0x0000: "i.MXRT595, i.MXRT685",
        0x0001: "LPC550x, LPC55s0x, LPC551x, LPC55s1x, LPC552x, LPC55s2x, LPC55s6",
        0x0004: "LPC55s3",
        0x0005: "KW45xx/K32W1xx",
        0x0006: "MCXN9xx (A0)",
        0x0007: "MCXN9xx (A1)",
        0x000A: "RW61x (A2)",
        0x5254049C: "i.MXRT118x",
    }

    def __init__(
        self,
        socc: int,
        uuid: bytes,
        rot_meta: bytes,
        dck_pub: bytes,
        cc_socu: int,
        cc_vu: int,
        cc_beacon: int,
        rot_pub: bytes,
        signature: Optional[bytes] = None,
        signature_provider: Optional[SignatureProvider] = None,
    ) -> None:
        """Initialize the DebugCredential object.

        :param socc: The SoC Class that this credential applies to
        :param uuid: The bytes of the unique device identifier
        :param rot_meta: Metadata for Root of Trust
        :param dck_pub: Internal binary representation of Debug Credential public key
        :param cc_socu: The Credential Constraint value that the vendor has associated with this credential.
        :param cc_vu: The Vendor Usage constraint value that the vendor has associated with this credential.
        :param cc_beacon: The non-zero Credential Beacon value, which is bound to a DC
        :param rot_pub: Internal binary representation of RoT public key
        :param signature: Debug Credential signature
        :param signature_provider: external signature provider
        """
        self.socc = socc
        self.uuid = uuid
        self.rot_meta = rot_meta
        self.dck_pub = dck_pub
        self.cc_socu = cc_socu
        self.cc_vu = cc_vu
        self.cc_beacon = cc_beacon
        self.rot_pub = rot_pub
        self.signature = signature
        self.signature_provider = signature_provider

    def export(self) -> bytes:
        """Export to binary form (serialization).

        :return: binary representation of the debug credential
        :raises SPSDKError: When Debug Credential Signature is not set, call the .sign method first
        """
        # make sure user called .sign before
        if not self.signature:
            raise SPSDKError("Debug Credential Signature is not set, call the .sign method first")
        data = pack(
            self.FORMAT,
            *[int(v) for v in self.VERSION.split(".")],
            self.socc,
            self.uuid,
            self.rot_meta,
            self.dck_pub,
            self.cc_socu,
            self.cc_vu,
            self.cc_beacon,
            self.rot_pub,
            self.signature,
        )
        return data

    @staticmethod
    def get_socc_description(version: str, socc: int) -> str:
        """Get SOCC family name description.

        :param version: Protocol version
        :param socc: SOCC number
        :return: SOCC string representation
        """
        cls = DebugCredential._get_class(version, socc)
        return f"{socc:{cls.SOCC_FORMAT}}, " + cls.SOCC_LIST.get(socc, "Unknown SOCC")

    def info(self) -> str:
        """String representation of DebugCredential.

        :return: binary representation of the debug credential
        """
        msg = f"Version : {self.VERSION}\n"
        msg += f"SOCC    : {self.get_socc_description(self.VERSION, self.socc)}\n"
        msg += f"UUID    : {self.uuid.hex().upper()}\n"
        msg += f"CC_SOCC : {hex(self.cc_socu)}\n"
        msg += f"CC_VU   : {hex(self.cc_vu)}\n"
        msg += f"BEACON  : {self.cc_beacon}\n"
        msg += f"RoTKH   : {self.get_rotkh().hex()}\n"
        return msg

    def sign(self) -> None:
        """Sign the DC data using SignatureProvider."""
        if not self.signature_provider:
            raise SPSDKError("Debug Credential Signature provider is not set")
        signature = self.signature_provider.sign(self._get_data_to_sign())
        if not signature:
            raise SPSDKError("Debug Credential Signature provider didn't return any signature")
        self.signature = signature

    def _get_data_to_sign(self) -> bytes:
        """Collects data meant for signing."""
        data = pack(
            self.FORMAT_NO_SIG,
            *[int(v) for v in self.VERSION.split(".")],
            self.socc,
            self.uuid,
            self.rot_meta,
            self.dck_pub,
            self.cc_socu,
            self.cc_vu,
            self.cc_beacon,
            self.rot_pub,
        )
        return data

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, DebugCredential) and vars(self) == vars(other)

    @classmethod
    def _get_rot_meta(cls, used_root_cert: int, rot_pub_keys: List[str]) -> bytes:
        """Creates the RoT meta-data required by the device to corroborate.

        :return: binary representing the rot-meta data
        :raises NotImplementedError: Derived class has to implement this method
        """
        raise NotImplementedError("Derived class has to implement this method.")

    @staticmethod
    def _get_dck(dck_key_path: str) -> bytes:
        """Loads the Debugger Public Key (DCK).

        :return: binary representing the DCK key
        :raises NotImplementedError: Derived class has to implement this method
        """
        raise NotImplementedError("Derived class has to implement this method.")

    @staticmethod
    def _get_rot_pub(rot_pub_id: int, rot_pub_keys: List[str]) -> bytes:
        """Loads the vendor RoT Public key that corresponds to the private key used for singing.

        :return: binary representing the rotk public key
        :raises NotImplementedError: Derived class has to implement this method
        """
        raise NotImplementedError("Derived class has to implement this method.")

    def get_rotkh(self) -> bytes:
        """Get Root Of Trust Keys Hash.

        :return: RoTKH in bytes
        :raises NotImplementedError: Derived class has to implement this method
        """
        raise NotImplementedError("Derived class has to implement this method.")

    @classmethod
    def _get_class(cls, version: str, socc: int) -> "Type[DebugCredential]":
        if socc in [0x5254049C]:
            return _edge_lock_version_mapping[version]
        return _version_mapping[version]

    @classmethod
    def create_from_yaml_config(
        cls, version: str, yaml_config: dict, search_paths: Optional[List[str]] = None
    ) -> "DebugCredential":
        """Create a debug credential object out of yaml configuration.

        :param version: Debug Authentication protocol version.
        :param yaml_config: Debug credential file configuration.
        :param search_paths: List of paths where to search for the file, defaults to None

        :return: DebugCredential object
        """
        socc = yaml_config["socc"]
        klass = DebugCredential._get_class(version=version, socc=socc)
        # Fix the file paths by search paths
        for i, rot in enumerate(yaml_config["rot_meta"]):
            yaml_config["rot_meta"][i] = find_file(rot, search_paths=search_paths)
        if "rotk" in yaml_config.keys():
            yaml_config["rotk"] = find_file(yaml_config["rotk"], search_paths=search_paths)
        yaml_config["dck"] = find_file(yaml_config["dck"], search_paths=search_paths)
        signature_provider = get_signature_provider(
            sp_cfg=yaml_config.get("sign_provider"),
            local_file_key=yaml_config.get("rotk"),
            search_paths=search_paths,
        )
        dc_obj = klass(
            socc=yaml_config["socc"],
            uuid=bytes.fromhex(yaml_config["uuid"]),
            rot_meta=klass._get_rot_meta(  # pylint: disable=protected-access
                used_root_cert=yaml_config["rot_id"],
                rot_pub_keys=yaml_config["rot_meta"],
            ),
            dck_pub=klass._get_dck(yaml_config["dck"]),  # pylint: disable=protected-access
            cc_socu=yaml_config["cc_socu"],
            cc_vu=yaml_config["cc_vu"],
            cc_beacon=yaml_config["cc_beacon"],
            rot_pub=klass._get_rot_pub(  # pylint: disable=protected-access
                yaml_config["rot_id"], yaml_config["rot_meta"]
            ),
            signature_provider=signature_provider,
        )
        return dc_obj

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> "DebugCredential":
        """Parse the debug credential.

        :param data: Raw data as bytes
        :param offset: Offset of input data
        :return: DebugCredential object
        """
        ver = unpack_from("<2H", data, offset)
        version = f"{ver[0]}.{ver[1]}"
        socc = unpack_from("<L", data, offset + 4)
        klass = cls._get_class(version, socc[0])
        return klass.get_instance_from_challenge(data[offset:])

    @classmethod
    def get_instance_from_challenge(cls, data: bytes) -> "DebugCredential":
        """Returns instance of class from DAP authentication challenge data.

        :return: Instance of this class.
        """
        _, _, *rest = unpack_from(cls.FORMAT, data, 0)
        return cls(*rest)


class DebugCredentialRSA(DebugCredential):
    """Class for RSA specific of DebugCredential."""

    FORMAT_NO_SIG = "<2HL16s128s260s3L260s"
    FORMAT = FORMAT_NO_SIG + "256s"

    @classmethod
    def _get_rot_meta(cls, used_root_cert: int, rot_pub_keys: List[str]) -> bytes:
        """Creates the RoT meta-data required by the device to corroborate.

        The meta-data is created by getting the public numbers (modulus and exponent)
        from each of the RoT public keys, hashing them and combing together.

        :return: binary representing the rot-meta data
        """
        rot_meta = bytearray(128)
        for index, rot_key in enumerate(rot_pub_keys):
            rot = extract_public_key(file_path=rot_key, password=None)
            assert isinstance(rot, crypto.RSAPublicKey)
            data = rsa_key_to_bytes(key=rot, exp_length=3, modulus_length=None)
            result = internal_backend.hash(data)
            rot_meta[index * 32 : (index + 1) * 32] = result
        return bytes(rot_meta)

    @staticmethod
    def _get_dck(dck_key_path: str) -> bytes:
        """Loads the Debugger Public Key (DCK).

        :return: binary representing the DCK key
        """
        dck_key = extract_public_key(file_path=dck_key_path)
        assert isinstance(dck_key, crypto.RSAPublicKey)
        return rsa_key_to_bytes(key=dck_key, exp_length=4)

    @staticmethod
    def _get_rot_pub(rot_pub_id: int, rot_pub_keys: List[str]) -> bytes:
        """Loads the vendor RoT private key.

         It corresponds to the (default) position zero RoT key in the rot_meta list of public keys.
         Derive public key from RoT private keys and converts it to the bytes.

        :return: binary representing the rotk public key
        """
        pub_key_path = rot_pub_keys[rot_pub_id]
        pub_key = extract_public_key(file_path=pub_key_path, password=None)
        assert isinstance(pub_key, crypto.RSAPublicKey)
        return rsa_key_to_bytes(key=pub_key, exp_length=4)

    def get_rotkh(self) -> bytes:
        """Get Root Of Trust Keys Hash.

        :return: RoTKH in bytes
        """
        return internal_backend.hash(data=self.rot_meta[:])


class DebugCredentialECC(DebugCredential):
    """Class for ECC specific of DebugCredential."""

    HASH_LENGTH = 0
    KEY_LENGTH = 0
    CORD_LENGTH = 0
    HASH_SIZES = {32: 256, 48: 384, 66: 512}
    CURVE: crypto.ec.EllipticCurve = crypto.ec.SECP256R1()

    def sign(self) -> None:
        """Sign the DC data using SignatureProvider."""
        super().sign()
        if not self.signature:
            raise SPSDKError("Debug Credential Signature is not set in base class")
        r, s = crypto.utils_cryptography.decode_dss_signature(self.signature)
        public_numbers = crypto.EllipticCurvePublicNumbers(r, s, self.CURVE)
        self.signature = ecc_public_numbers_to_bytes(
            public_numbers=public_numbers, length=self.CORD_LENGTH
        )

    @classmethod
    def _get_rot_meta(cls, used_root_cert: int, rot_pub_keys: List[str]) -> bytes:
        """Creates the RoT meta-data required by the device to corroborate.

        :return: binary representing the rot-meta data
        """
        ctrk_hash_table = DebugCredentialECC.create_ctrk_table(rot_pub_keys)
        flags = DebugCredentialECC.calculate_flags(used_root_cert, rot_pub_keys)
        return flags + ctrk_hash_table

    @staticmethod
    def _get_dck(dck_key_path: str) -> bytes:
        """Loads the Debugger Public Key (DCK).

        :return: binary representing the DCK key
        """
        dck_key = extract_public_key(file_path=dck_key_path)
        length = math.ceil(dck_key.key_size / 8)
        assert isinstance(dck_key, crypto.EllipticCurvePublicKey)
        data = ecc_key_to_bytes(dck_key, length=length)
        return data

    @staticmethod
    def _get_rot_pub(rot_pub_id: int, rot_pub_keys: List[str]) -> bytes:
        """Loads the vendor RoT Public key that corresponds to the private key used for singing.

        :return: binary representing the rotk public key
        """
        root_key = rot_pub_keys[rot_pub_id]
        root_public_key = extract_public_key(file_path=root_key, password=None)
        length = math.ceil(root_public_key.key_size / 8)
        assert isinstance(root_public_key, crypto.EllipticCurvePublicKey)
        data = ecc_key_to_bytes(root_public_key, length=length)
        return data

    def info(self) -> str:
        """String representation of DebugCredential.

        :return: binary representation of the debug credential
        """
        msg = f"Version : {self.VERSION}\n"
        msg += f"SOCC    : {self.get_socc_description(self.VERSION, self.socc)}\n"
        msg += f"UUID    : {self.uuid.hex().upper()}\n"
        msg += f"CC_SOCC : {hex(self.cc_socu)}\n"
        msg += f"CC_VU   : {hex(self.cc_vu)}\n"
        msg += f"BEACON  : {self.cc_beacon}\n"
        ctrk_records_num = self.rot_meta[0] >> 4
        if ctrk_records_num == 1:
            msg += "CRTK table not present \n"
        else:
            msg += f"CRTK table has {ctrk_records_num} entries\n"
            msg += f"CRTK Hash: {self.get_rotkh().hex()}"
        return msg

    @property
    def FORMAT(self) -> str:  # type: ignore # pylint: disable=invalid-name
        """Formatting string."""
        return f"<2HL16s3L{len(self.rot_meta)}s{self.HASH_LENGTH * 2}s{self.HASH_LENGTH * 2}s{self.HASH_LENGTH * 2}s"

    @property
    def FORMAT_NO_SIG(self) -> str:  # type: ignore # pylint: disable=invalid-name
        """Formatting string without signature."""
        return f"<2HL16s3L{len(self.rot_meta)}s{self.HASH_LENGTH * 2}s{self.HASH_LENGTH * 2}s"

    @staticmethod
    def create_ctrk_table(rot_pub_keys: List[str]) -> bytes:
        """Creates ctrk table."""
        if len(rot_pub_keys) == 1:
            return bytes()
        ctrk_table = bytes()
        for pub_key_path in rot_pub_keys:
            pub_key = extract_public_key(file_path=pub_key_path, password=None)
            assert isinstance(pub_key, crypto.EllipticCurvePublicKey)
            hash_size = DebugCredentialECC.HASH_SIZES[math.ceil(pub_key.key_size / 8)]
            data = ecc_key_to_bytes(key=pub_key, length=math.ceil(pub_key.key_size / 8))
            ctrk_hash = internal_backend.hash(data=data, algorithm=f"sha{hash_size}")
            ctrk_table += ctrk_hash
        return ctrk_table

    @staticmethod
    def calculate_flags(used_root_cert: int, rot_pub_keys: List[str]) -> bytes:
        """Calculates flags in rotmeta."""
        flags = 0
        flags |= 1 << 31
        flags |= used_root_cert << 8
        flags |= len(rot_pub_keys) << 4
        return pack("<L", flags)

    def export(self) -> bytes:
        """Export to binary form (serialization)."""
        data = pack(
            self.FORMAT,
            *[int(v) for v in self.VERSION.split(".")],
            self.socc,
            self.uuid,
            self.cc_socu,
            self.cc_vu,
            self.cc_beacon,
            self.rot_meta,
            self.rot_pub,
            self.dck_pub,
            self.signature,
        )
        return data

    def _get_data_to_sign(self) -> bytes:
        """Collects data meant for signing."""
        data = pack(
            self.FORMAT_NO_SIG,
            *[int(v) for v in self.VERSION.split(".")],
            self.socc,
            self.uuid,
            self.cc_socu,
            self.cc_vu,
            self.cc_beacon,
            self.rot_meta,
            self.rot_pub,
            self.dck_pub,
        )
        return data

    def __eq__(self, other: Any) -> bool:
        self_vars = vars(self)
        del self_vars["signature_provider"]
        other_vars = vars(other)
        del other_vars["signature_provider"]
        return isinstance(other, DebugCredential) and other_vars == self_vars

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> "DebugCredential":
        """Parse the debug credential.

        :param data: Raw data as bytes
        :param offset: Offset of input data
        :return: DebugCredential object
        :raises SPSDKError: When flag is invalid
        """
        format_head = "<2HL16s4L"
        (
            version_major,  # pylint: disable=unused-variable
            version_minor,  # pylint: disable=unused-variable
            socc,
            uuid,
            cc_socu,
            cc_vu,
            beacon,
            flags,
        ) = unpack_from(format_head, data)
        if not flags & 0x8000_0000:
            raise SPSDKError("Invalid flag")
        records_num = (flags & 0xF0) >> 4
        rot_meta_len = 4
        ctrk_hash_table = bytes()
        if records_num > 1:
            rot_meta_len += records_num * cls.HASH_LENGTH
            ctrk_format = f"<{records_num * cls.HASH_LENGTH}s"
            ctrk_hash_table = unpack_from(ctrk_format, data, offset=offset + calcsize(format_head))[
                0
            ]
        rot_meta = pack("<L", flags) + ctrk_hash_table
        format_tail = f"<{cls.HASH_LENGTH * 2}s{cls.HASH_LENGTH * 2}s{cls.HASH_LENGTH * 2}s"
        rot_pub, dck_pub, signature = unpack_from(
            format_tail, data, offset + calcsize(format_head) + len(rot_meta) - 4
        )

        return cls(
            socc=socc,
            uuid=uuid,
            rot_meta=rot_meta,
            dck_pub=dck_pub,
            cc_socu=cc_socu,
            cc_vu=cc_vu,
            cc_beacon=beacon,
            rot_pub=rot_pub,
            signature=signature,
        )

    def get_rotkh(self) -> bytes:
        """Get Root Of Trust Keys Hash.

        :return: RoTKH in bytes
        """
        srk_records_num = self.rot_meta[0] >> 4
        if srk_records_num == 1:
            return b""
        key_length = 256 if ((len(self.rot_meta) - 4) // srk_records_num) == 32 else 384
        return internal_backend.hash(data=self.rot_meta[4:], algorithm=f"sha{key_length}")

    @classmethod
    def get_instance_from_challenge(cls, data: bytes) -> "DebugCredential":
        """Returns instance of class from DAP authentication challenge data.

        :return: Instance of this class.
        """
        return cls.parse(data, 0)


class DebugCredentialEdgeLockEnclave(DebugCredentialECC):
    """EdgeLock Class."""

    HASH_LENGTH = 0
    KEY_LENGTH = 0
    CORD_LENGTH = 0
    SOCC_FORMAT = "08X"

    @classmethod
    def _get_rot_meta(cls, used_root_cert: int, rot_pub_keys: List[str]) -> bytes:
        """Creates the RoT meta-data required by the device to corroborate.

        :return: binary representing the rot-meta data
        """
        srk_hash_table = DebugCredentialEdgeLockEnclave.create_srk_table(rot_pub_keys)
        flags = DebugCredentialECC.calculate_flags(used_root_cert, rot_pub_keys)
        return flags + srk_hash_table

    @staticmethod
    def _get_dck(dck_key_path: str) -> bytes:
        """Loads the Debugger Public Key (DCK).

        :return: binary representing the DCK key
        """
        dck_key = extract_public_key(file_path=dck_key_path)
        length = math.ceil(dck_key.key_size / 8)
        assert isinstance(dck_key, crypto.EllipticCurvePublicKey)
        data = ecc_key_to_bytes(dck_key, length=length)
        return data

    @staticmethod
    def _get_rot_pub(rot_pub_id: int, rot_pub_keys: List[str]) -> bytes:
        """Loads the vendor RoT Public key that corresponds to the private key used for singing.

        :return: binary representing the rotk public key
        """
        return DebugCredentialECC._get_rot_pub(rot_pub_id=rot_pub_id, rot_pub_keys=rot_pub_keys)

    def get_rotkh(self) -> bytes:
        """Get Root Of Trust Keys Hash.

        :return: RoTKH in bytes
        """
        srk = SRKTable.parse(self.rot_meta[4:])
        srk.update_fields()
        return srk.compute_srk_hash()

    def info(self) -> str:
        """String representation of DebugCredential.

        :return: binary representation of the debug credential
        """
        msg = f"Version : {self.VERSION}\n"
        msg += f"SOCC    : {self.get_socc_description(self.VERSION, self.socc)}\n"
        msg += f"UUID    : {self.uuid.hex().upper()}\n"
        msg += f"CC_SOCC : {hex(self.cc_socu)}\n"
        msg += f"CC_VU   : {hex(self.cc_vu)}\n"
        msg += f"BEACON  : {self.cc_beacon}\n"
        srk_records_num = self.rot_meta[0] >> 4
        if srk_records_num != 4:
            msg += "Invalid count of SRK records \n"
        else:
            msg += f"SRK table has {srk_records_num} entries\n"
            msg += f"SRK Hash: {self.get_rotkh().hex()}"
        return msg

    @property
    def FORMAT(self) -> str:  # type: ignore # pylint: disable=invalid-name
        """Formatting string."""
        return f"<2HL16s3L{len(self.rot_meta)}s{self.HASH_LENGTH * 2}s{self.HASH_LENGTH * 2}s"

    @property
    def FORMAT_NO_SIG(self) -> str:  # type: ignore # pylint: disable=invalid-name
        """Formatting string without signature."""
        return f"<2HL16s3L{len(self.rot_meta)}s{self.HASH_LENGTH * 2}s"

    @staticmethod
    def create_srk_table(rot_pub_keys: List[str]) -> bytes:
        """Creates ctrk table."""
        if len(rot_pub_keys) != 4:
            raise SPSDKValueError("Invalid count of Super Root keys!")

        srk_table = SRKTable(
            [SRKRecord.create_from_key(extract_public_key(x)) for x in rot_pub_keys]
        )
        srk_table.update_fields()
        srk_table.validate({})
        return srk_table.export()

    def export(self) -> bytes:
        """Export to binary form (serialization)."""
        data = pack(
            self.FORMAT,
            *[int(v) for v in self.VERSION.split(".")],
            self.socc,
            self.uuid,
            self.cc_socu,
            self.cc_vu,
            self.cc_beacon,
            self.rot_meta,
            self.dck_pub,
            self.signature,
        )
        return data

    def _get_data_to_sign(self) -> bytes:
        """Collects data meant for signing."""
        data = pack(
            self.FORMAT_NO_SIG,
            *[int(v) for v in self.VERSION.split(".")],
            self.socc,
            self.uuid,
            self.cc_socu,
            self.cc_vu,
            self.cc_beacon,
            self.rot_meta,
            self.dck_pub,
        )
        return data

    def __eq__(self, other: Any) -> bool:
        self_vars = vars(self)
        del self_vars["signature_provider"]
        other_vars = vars(other)
        del other_vars["signature_provider"]
        return isinstance(other, DebugCredential) and other_vars == self_vars

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> "DebugCredential":
        """Parse the debug credential.

        :param data: Raw data as bytes
        :param offset: Offset of input data
        :return: DebugCredential object
        :raises SPSDKError: When flag is invalid
        """
        format_head = "<2HL16s4L"
        (
            version_major,  # pylint: disable=unused-variable
            version_minor,  # pylint: disable=unused-variable
            socc,
            uuid,
            cc_socu,
            cc_vu,
            beacon,
            flags,
        ) = unpack_from(format_head, data)
        if not flags & 0x8000_0000:
            raise SPSDKError("Invalid flag")

        srk_table = SRKTable.parse(data, offset=offset + calcsize(format_head))
        srk_table.update_fields()
        srk_table.validate({})
        rot_meta = int.to_bytes(flags, 4, "little") + srk_table.export()
        format_tail = f"<{cls.HASH_LENGTH * 2}s{cls.HASH_LENGTH * 2}s"
        dck_pub, signature = unpack_from(
            format_tail, data, offset + calcsize(format_head) + len(srk_table)
        )

        return cls(
            socc=socc,
            uuid=uuid,
            rot_meta=rot_meta,
            dck_pub=dck_pub,
            rot_pub=bytes(),
            cc_socu=cc_socu,
            cc_vu=cc_vu,
            cc_beacon=beacon,
            signature=signature,
        )

    @classmethod
    def get_instance_from_challenge(cls, data: bytes) -> "DebugCredential":
        """Returns instance of class from DAP authentication challenge data.

        :return: Instance of this class.
        """
        return cls.parse(data, 0)


class DebugCredentialRSA2048(DebugCredentialRSA):
    """DebugCredential class for RSA 2048."""

    FORMAT_NO_SIG = "<2HL16s128s260s3L260s"
    FORMAT = FORMAT_NO_SIG + "256s"
    VERSION = "1.0"


class DebugCredentialRSA4096(DebugCredentialRSA):
    """DebugCredential class for RSA 4096."""

    FORMAT_NO_SIG = "<2HL16s128s516s3L516s"
    FORMAT = FORMAT_NO_SIG + "512s"
    VERSION = "1.1"


class DebugCredentialECC256(DebugCredentialECC):
    """DebugCredential class for LPC55s3x for version 2.0 (p256)."""

    VERSION = "2.0"
    CURVE = crypto.ec.SECP256R1()
    HASH_LENGTH = 32
    CORD_LENGTH = 32
    KEY_LENGTH = 256


class DebugCredentialECC384(DebugCredentialECC):
    """DebugCredential class for LPC55s3x for version 2.1 (p384)."""

    VERSION = "2.1"
    CURVE = crypto.ec.SECP384R1()
    HASH_LENGTH = 48
    CORD_LENGTH = 48
    KEY_LENGTH = 384


class DebugCredentialECC521(DebugCredentialECC):
    """DebugCredential class for LPC55s3x for version 2.1 (p384)."""

    VERSION = "2.2"
    CURVE = crypto.ec.SECP521R1()
    HASH_LENGTH = 66
    CORD_LENGTH = 66
    KEY_LENGTH = 521


class DebugCredentialEdgeLockEnclaveECC256(DebugCredentialEdgeLockEnclave):
    """Debug Credential class for device using EdgeLock peripheral for ECC256 keys."""

    VERSION = "2.0"
    CURVE = crypto.ec.SECP256R1()
    HASH_LENGTH = 32
    CORD_LENGTH = 32
    KEY_LENGTH = 256


class DebugCredentialEdgeLockEnclaveECC384(DebugCredentialEdgeLockEnclave):
    """Debug Credential class for device using EdgeLock peripheral for ECC384 keys."""

    VERSION = "2.1"
    CURVE = crypto.ec.SECP384R1()
    HASH_LENGTH = 48
    CORD_LENGTH = 48
    KEY_LENGTH = 384


class DebugCredentialEdgeLockEnclaveECC521(DebugCredentialEdgeLockEnclave):
    """Debug Credential class for device using EdgeLock peripheral for ECC521 keys."""

    VERSION = "2.2"
    CURVE = crypto.ec.SECP521R1()
    HASH_LENGTH = 66
    CORD_LENGTH = 66
    KEY_LENGTH = 521


_version_mapping = {
    "1.0": DebugCredentialRSA2048,
    "1.1": DebugCredentialRSA4096,
    "2.0": DebugCredentialECC256,
    "2.1": DebugCredentialECC384,
    "2.2": DebugCredentialECC521,
}


_edge_lock_version_mapping = {
    "2.0": DebugCredentialEdgeLockEnclaveECC256,
    "2.1": DebugCredentialEdgeLockEnclaveECC384,
    "2.2": DebugCredentialEdgeLockEnclaveECC521,
}



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2022-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause
"""DK6 Device Commands."""
from struct import unpack_from
from typing import Dict, Type, Union

from spsdk.utils.easy_enum import Enum


class StatusCode(Enum):
    """DK6 Command Status Codes."""

    OK = (0x00, "Success")
    MEMORY_INVALID_MODE = (0xEF, "Memory invalid mode")
    MEMORY_BAD_STATE = (0xF0, "Memory bad state")
    MEMORY_TOO_LONG = (0xF1, "Memory too long")
    MEMORY_OUT_OF_RANGE = (0xF2, "Memory out of range")
    MEMORY_ACCESS_INVALID = (0xF3, "Memory access invalid")
    MEMORY_NOT_SUPPORTED = (0xF4, "Memory not supported")
    MEMORY_INVALID = (0xF5, "Memory invalid")
    NO_RESPONSE = (0xF6, "No response")
    AUTH_ERROR = (0xF7, "Not authorized")
    TEST_ERROR = (0xF8, "Test Error")
    READ_FAIL = (0xF9, "Read fail")
    USER_INTERRUPT = (0xFA, "User interrupt")
    ASSERT_FAIL = (0xFB, "Assertion fail")
    CRC_ERROR = (0xFC, "CRC Error")
    INVALID_RESPONSE = (0xFD, "Invalid response")
    WRITE_FAIL = (0xFE, "Write fail")
    NOT_SUPPORTED = (0xFF, "Not supported")


class ResponseTag(Enum):
    """DK6 Responses to Commands."""

    RESET = (0x15, "ResetResponse", "Reset Response")
    EXECUTE = (0x22, "ExecuteResponse", "Execute Response")
    SET_BAUD = (0x28, "SetBaudResponse", "Set Baud Response")
    GET_CHIPID = (0x33, "GetChipIdResponse", "Get Chip ID Response")
    MEM_OPEN = (0x41, "MemOpenResponse", "Memory Open Response")
    MEM_ERASE = (0x43, "MemEraseResponse", "Memory Erase Response")
    MEM_BLANK_CHECK = (0x45, "MemBlankCheckResponse", "Memory Blank Check Response")
    MEM_READ = (0x47, "MemReadResponse", "Memory Read Response")
    MEM_WRITE = (0x49, "MemoryWriteResponse", "Memory Write Response")
    MEM_CLOSE = (0x4B, "MemoryCloseResponse", "Memory Close Response")
    MEM_GET_INFO = (0x4D, "MemoryGetInfoResponse", "Memory Get Info Response")
    UNLOCK_ISP = (0x4F, "UnlockISPResponse", "Unlock ISP Response")


class CommandTag(Enum):
    """DK6 Commands."""

    RESET = (0x14, "ResetCommand", "Reset Command")
    EXECUTE = (0x21, "ExecuteCommand", "Execute Command")
    SET_BAUD = (0x27, "SetBaudCommand", "Set Baud Command")
    GET_CHIPID = (0x32, "GetChipIdCommand", "Get Chip ID Command")
    MEM_OPEN = (0x40, "MemOpenCommand", "Memory Open Command")
    MEM_ERASE = (0x42, "MemEraseCommand", "Memory Erase Command")
    MEM_BLANK_CHECK = (0x44, "MemBlankCheckCommand", "Memory Blank Check Command")
    MEM_READ = (0x46, "MemReadCommand", "Memory Read Command")
    MEM_WRITE = (0x48, "MemoryWriteCommand", "Memory Write Command")
    MEM_CLOSE = (0x4A, "MemoryCloseCommand", "Memory Close Command")
    MEM_GET_INFO = (0x4C, "MemoryGetInfoCommand", "Memory Get Info Command")
    UNLOCK_ISP = (0x4E, "UnlockISPCommand", "Unlock ISP Command")


class MemoryId(Enum):
    """DK6 Memory IDs."""

    FLASH = (0x00, "FLASH")
    PSECT = (0x01, "PSECT")
    pFLASH = (0x02, "pFLASH")
    Config = (0x03, "Config")
    EFUSE = (0x04, "EFUSE")
    ROM = (0x05, "ROM")
    RAM0 = (0x06, "RAM0")
    RAM1 = (0x07, "RAM1")


class MemoryType(Enum):
    """DK6 Memory Types."""

    ROM = (0x00, "ROM", "Read only memory")
    FLASH = (0x01, "FLASH", "FLASH memory")
    RAM = (0x02, "RAM", "RAM")
    EFUSE = (0x05, "EFUSE (OTP)", "EFUSE (OTP)")


class MemoryAccessValues(Enum):
    """DK6 Memory Access Values."""

    READ = (0x00, "Read Enabled")
    WRITE = (0x01, "Write Enabled")
    ERASE = (0x02, "Erase Enabled")
    ERASE_ALL = (0x03, "Erase All Enabled")
    BLANK_CHECK_ENABLED = (0x04, "Blank Check Enabled")
    ALL = (0x0F, "All is available")


class CmdPacket:
    """DK6 command packet format class."""

    def __init__(self, data: bytes) -> None:
        """Initialize the Command Packet object.

        :param data: Command data, defaults to None
        """
        self.data = data

    def __eq__(self, obj: object) -> bool:
        return isinstance(obj, CmdPacket) and vars(obj) == vars(self)

    def __ne__(self, obj: object) -> bool:
        return not self.__eq__(obj)

    def __str__(self) -> str:
        return "<" + self.info() + ">"

    def info(self) -> str:
        """Get object info."""
        return "CMDPacket[" + ", ".join(f"{b:02X}" for b in self.data) + "]"

    def to_bytes(self) -> Union[bytes, None]:
        """Serialize CmdPacket into bytes.

        :return: Serialized object into bytes
        """
        return self.data


class CmdResponse:
    """DK6 response base format class."""

    MSG_OFFSET = 1

    def __init__(self, cmd_type: int, raw_data: bytes) -> None:
        """Initialize the Command Response object.

        :param header: Header for the response
        :param raw_data: Response data
        """
        assert isinstance(raw_data, (bytes, bytearray))
        self.type = cmd_type
        self.status = raw_data[0]
        self.raw_data = raw_data

    def __eq__(self, obj: object) -> bool:
        return isinstance(obj, CmdResponse) and vars(obj) == vars(self)

    def __ne__(self, obj: object) -> bool:
        return not self.__eq__(obj)

    def __str__(self) -> str:
        return "<" + self.info() + ">"

    def info(self) -> str:
        """Get object info."""
        return (
            f"Status=0x{self.status:02X}"
            + f"Type=0x{self.type:02X}"
            + " ["
            + ", ".join(f"{b:02X}" for b in self.raw_data)
            + "]"
        )


class GenericResponse(CmdResponse):
    """DK6 generic response format class."""

    def info(self) -> str:
        """Get object info."""
        tag = ResponseTag.name(self.type)
        status = StatusCode.get(self.status, f"Unknown[0x{self.status:08X}]")
        return f"Tag={tag}, Status={status}"


class IspUnlockResponse(CmdResponse):
    """ISP Unlock response format class."""

    def __init__(self, cmd_type: int, raw_data: bytes) -> None:
        """Initialize the Command Response object.

        :param header: Header for the response
        :param raw_data: Response data
        """
        super().__init__(cmd_type, raw_data)
        self.authenticated = self.status == StatusCode.OK

    def info(self) -> str:
        """Get object info."""
        tag = ResponseTag.name(self.type)
        status = StatusCode.get(self.status, f"Unknown[0x{self.status:08X}]")
        return f"Tag={tag}, Status={status}, Authenticated={self.authenticated}"


class GetChipIdResponse(CmdResponse):
    """Chip get info response format class."""

    FORMAT = "<II"

    def __init__(self, cmd_type: int, raw_data: bytes) -> None:
        """Initialize the Flash-Read-Once response object.

        :param header: Header for the response
        :param raw_data: Response data
        """
        super().__init__(cmd_type, raw_data)
        if self.status == StatusCode.OK:
            (self.chip_id, self.chip_version) = unpack_from(self.FORMAT, raw_data, self.MSG_OFFSET)

    def info(self) -> str:
        """Get object info."""
        tag = ResponseTag.name(self.type)
        status = StatusCode.get(self.status, f"Unknown[0x{self.status:08X}]")
        if self.status == StatusCode.OK:
            return f"Tag={tag}, Status={status}, ChipID={hex(self.chip_id)}, ChipVersion={hex(self.chip_version)}"
        return f"Tag={tag}, Status={status}"


class MemGetInfoResponse(CmdResponse):
    """Memory get info response format class."""

    FORMAT = "<BIIIBB"
    MEM_NAME_OFFSET = 15

    def __init__(self, cmd_type: int, raw_data: bytes) -> None:
        """Initialize the Flash-Read-Once response object.

        :param header: Header for the response
        :param raw_data: Response data
        """
        super().__init__(cmd_type, raw_data)
        if self.status == StatusCode.OK:
            self.mem_name = raw_data[self.MEM_NAME_OFFSET :].decode("ascii")
            (
                self.memory_id,
                self.base_addr,
                self.length,
                self.sector_size,
                self.mem_type,
                self.access,
            ) = unpack_from(self.FORMAT, raw_data, self.MSG_OFFSET)

    def info(self) -> str:
        """Get object info."""
        tag = ResponseTag.name(self.type)
        status = StatusCode.get(self.status, f"Unknown[0x{self.status:08X}]")
        if self.status == StatusCode.OK:
            memory_id = MemoryId.get(self.memory_id)
            mem_type = MemoryType.get(self.mem_type)
            access = MemoryAccessValues.get(self.access)
            return (
                f"Tag={tag}, Status={status}, MemName = {self.mem_name}, "
                + f"MemoryId={memory_id}, BaseAddress={hex(self.base_addr)}, "
                + f"Length={hex(self.length)}, SectorSize={hex(self.sector_size)}, "
                + f"MemoryType={mem_type}, Access={access}"
            )
        return f"Tag={tag}, Status={status}"


class MemOpenResponse(CmdResponse):
    """Memory open response format class."""

    FORMAT = "<B"

    def __init__(self, cmd_type: int, raw_data: bytes) -> None:
        """Initialize the Command Response object.

        :param header: Header for the response
        :param raw_data: Response data
        """
        super().__init__(cmd_type, raw_data)
        if self.status == StatusCode.OK:
            self.handle = unpack_from(self.FORMAT, raw_data, self.MSG_OFFSET)

    def info(self) -> str:
        """Get object info."""
        tag = ResponseTag.name(self.type)
        status = StatusCode.get(self.status, f"Unknown[0x{self.status:08X}]")
        if self.status == StatusCode.OK:
            return f"Tag={tag}, Status={status}, Handle={self.handle}"
        return f"Tag={tag}, Status={status}"


class MemReadResponse(CmdResponse):
    """Memory open response format class."""

    def __init__(self, cmd_type: int, raw_data: bytes) -> None:
        """Initialize the Command Response object.

        :param header: Header for the response
        :param raw_data: Response data
        """
        super().__init__(cmd_type, raw_data)
        self.data = b""
        if self.status == StatusCode.OK:
            self.data = self.raw_data[self.MSG_OFFSET :]

    def info(self) -> str:
        """Get object info."""
        tag = ResponseTag.name(self.type)
        status = StatusCode.get(self.status, f"Unknown[0x{self.status:08X}]")
        data = ", ".join(f"{b:02X}" for b in self.data)
        return f"Tag={tag}, Status={status}, Data={data}"


class MemWriteResponse(CmdResponse):
    """Memory open response format class."""

    def info(self) -> str:
        """Get object info."""
        tag = ResponseTag.name(self.type)
        status = StatusCode.get(self.status, f"Unknown[0x{self.status:08X}]")
        return f"Tag={tag}, Status={status}"


class MemEraseResponse(CmdResponse):
    """Memory open response format class."""

    def info(self) -> str:
        """Get object info."""
        tag = ResponseTag.name(self.type)
        status = StatusCode.get(self.status, f"Unknown[0x{self.status:08X}]")
        return f"Tag={tag}, Status={status}"


class MemBlankCheckResponse(CmdResponse):
    """Memory open response format class."""

    def info(self) -> str:
        """Get object info."""
        tag = ResponseTag.name(self.type)
        status = StatusCode.get(self.status, f"Unknown[0x{self.status:08X}]")
        return f"Tag={tag}, Status={status}"


class MemCloseResponse(CmdResponse):
    """DK6 memory close response format class."""

    def info(self) -> str:
        """Get object info."""
        tag = ResponseTag.name(self.type)
        status = StatusCode.get(self.status, f"Unknown[0x{self.status:08X}]")
        return f"Tag={tag}, Status={status}"


def parse_cmd_response(
    data: bytes, frame_type: int
) -> Union[
    CmdResponse,
    GenericResponse,
    GetChipIdResponse,
    MemOpenResponse,
    MemEraseResponse,
    MemBlankCheckResponse,
    MemReadResponse,
    MemWriteResponse,
    MemCloseResponse,
    MemGetInfoResponse,
    IspUnlockResponse,
]:
    """Parse command response.

    :param data: Input data in bytes
    :param frame_type: Frame Type
    :return: De-serialized object from data
    """
    known_response: Dict[int, Type[CmdResponse]] = {
        ResponseTag.RESET: GenericResponse,
        ResponseTag.EXECUTE: GenericResponse,
        ResponseTag.SET_BAUD: GenericResponse,
        ResponseTag.GET_CHIPID: GetChipIdResponse,
        ResponseTag.MEM_OPEN: MemOpenResponse,
        ResponseTag.MEM_ERASE: MemEraseResponse,
        ResponseTag.MEM_BLANK_CHECK: MemBlankCheckResponse,
        ResponseTag.MEM_READ: MemReadResponse,
        ResponseTag.MEM_WRITE: MemWriteResponse,
        ResponseTag.MEM_CLOSE: MemCloseResponse,
        ResponseTag.MEM_GET_INFO: MemGetInfoResponse,
        ResponseTag.UNLOCK_ISP: IspUnlockResponse,
    }

    if frame_type in known_response:
        return known_response[frame_type](frame_type, data)
    return CmdResponse(frame_type, data)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2021-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause
"""Trust provisioning - TP Target, ISP mode over BLHOST."""

from enum import Enum
from typing import Any, Dict, List, Optional, Sequence

from spsdk.mboot import interfaces
from spsdk.mboot.exceptions import McuBootError, StatusCode
from spsdk.mboot.interfaces.base import MBootInterface
from spsdk.mboot.interfaces.uart import Uart
from spsdk.mboot.interfaces.usb import RawHid
from spsdk.mboot.mcuboot import McuBoot
from spsdk.tp import TP_SCH_FILE, TpTargetInterface
from spsdk.tp.exceptions import SPSDKTpTargetError
from spsdk.tp.tp_intf import TpIntfDescription
from spsdk.utils.misc import value_to_int
from spsdk.utils.schema_validator import ValidationSchemas


class TpBlHostIntfDescription(TpIntfDescription):
    """TP BLHOST Interface description."""

    def __init__(
        self,
        name: str,
        description: str,
        settings: Optional[Dict],
    ) -> None:
        """The BLHOST adapter for TPHOST interface description class.

        :param name: Name of the target
        :param description: Description of the target
        :param settings: Settings of target
        """
        super().__init__(name, TpTargetBlHost, description, settings)
        self.device: Optional[MBootInterface] = None

    def as_dict(self) -> Dict[str, Any]:
        """Returns dictionary with important fields for selection table."""
        ret = {}
        ret["name"] = self.name
        ret["description"] = self.description
        # For USB device
        if isinstance(self.device, RawHid):
            ret["path_hash"] = self.device.path_hash
            ret["pid_vid"] = f"{self.device.vid:#06x}:{self.device.pid:#06x}"
        if isinstance(self.device, Uart):
            ret["port"] = self.device.device.port
            ret["baudrate"] = self.device.device.baudrate

        return ret

    def get_id(self) -> str:
        """Returns the ID of the interface (com port or VID:PID)."""
        if isinstance(self.device, Uart):
            return self.device.device.port
        if isinstance(self.device, RawHid):
            return f"{self.device.vid:#06x}:{self.device.pid:#06x}"
        raise SPSDKTpTargetError(f"Unknown target device type: {type(self.device)}")

    def get_id_hash(self) -> str:
        """Return the ID hash of the interface. (COM port or hash of USB path)."""
        if isinstance(self.device, Uart):
            return self.device.device.port
        if isinstance(self.device, RawHid):
            return self.device.path_hash
        raise SPSDKTpTargetError(f"Unknown target device type: {type(self.device)}")


class TpTargetBlHost(TpTargetInterface):
    """Trust provisioning target adapter for ISP mode over BLHOST."""

    NAME = "blhost"

    class SettingsKey(str, Enum):
        """Keys used in `get_connected_devices` in `settings` dictionary."""

        PORT = "blhost_port"
        USB = "blhost_usb"
        TIMEOUT = "blhost_timeout"
        BAUDRATE = "blhost_baudrate"

    @staticmethod
    def _get_settings(settings: Optional[Dict] = None) -> Dict:
        """The function gets the important parameters for BLHOST from general settings.

        :param settings: General TPHOST target settings, defaults to None
        :return: Just BLHOST adapter important settings.
        """
        ret = {}
        ret["usb"] = settings.get("blhost_usb", None) if settings else None
        ret["port"] = settings.get("blhost_port", None) if settings else None
        ret["baudrate"] = value_to_int(settings.get("blhost_baudrate", 57600)) if settings else None
        ret["timeout"] = value_to_int(settings.get("blhost_timeout", 50)) if settings else 50
        return ret

    @classmethod
    def get_connected_targets(cls, settings: Optional[Dict] = None) -> List[TpIntfDescription]:
        """Get all connected TP targets of this adapter.

        :param settings: Possible settings to determine the way to find connected device, defaults to None.
        :return: List of all founded TP targets.
        """
        ret: List[TpIntfDescription] = []
        desc = cls._get_settings(settings)

        handle_usb = not desc["port"]
        handle_uart = not desc["usb"]
        # special case, when someone defines both 'port' and 'usb'
        if desc["port"] and desc["usb"]:
            handle_uart, handle_usb = True, True

        if handle_usb:
            usb_targets: Sequence[RawHid] = interfaces.scan_usb(desc["usb"])
            for usb_target in usb_targets:
                name = usb_target.name
                description = "BLHOST USB target " + str(usb_target.interface_number)
                usbt_desc = TpBlHostIntfDescription(name, description, settings)
                usbt_desc.device = usb_target
                usbt_desc.device.timeout = desc["timeout"]
                ret.append(usbt_desc)

        if handle_uart:
            uart_targets: Sequence[Uart] = interfaces.scan_uart(
                desc["port"], desc["baudrate"], desc["timeout"]
            )

            for uart_target in uart_targets:
                name = uart_target.device.port
                description = "BLHOST UART target"
                uart_desc = TpBlHostIntfDescription(name, description, settings)
                uart_desc.device = uart_target
                ret.append(uart_desc)

        return ret

    get_connected_interfaces = get_connected_targets

    def __init__(self, descriptor: TpBlHostIntfDescription) -> None:
        """Initialization of provisioned device adapter.

        :param descriptor: BLHOST adapter interface description.
        :raises SPSDKTpTargetError: None existing device.
        """
        super().__init__(descriptor=descriptor)
        if not descriptor.device:
            raise SPSDKTpTargetError("Device is not defined.")
        self.mboot = McuBoot(descriptor.device)

        # TODO Get the right pseudo index instead of memory address
        self.buffer_address = (
            value_to_int(self.descriptor.settings.get("buffer_address", 0))
            if self.descriptor.settings
            else 0
        )
        # TODO Get the right buffer size
        self.buffer_size = (
            value_to_int(self.descriptor.settings.get("buffer_size", 0x1000))
            if self.descriptor.settings
            else 0x1000
        )

    @property
    def uses_uart(self) -> bool:
        """Check if the adapter is using UART for communication."""
        assert isinstance(self.descriptor, TpBlHostIntfDescription)
        return isinstance(self.descriptor.device, Uart)

    @property
    def uses_usb(self) -> bool:
        """Check if the adapter is using USB for communication."""
        assert isinstance(self.descriptor, TpBlHostIntfDescription)
        return isinstance(self.descriptor.device, RawHid)

    def open(self) -> None:
        """Open the provisioned device adapter."""
        self.mboot.open()
        self.mboot.reopen = True

    @property
    def is_open(self) -> bool:
        """Check if provisioned device adapter is open."""
        return self.mboot._device.is_opened

    def close(self) -> None:
        """Close the provisioned device adapter."""
        self.mboot.close()

    def reset_device(self) -> None:
        """Reset the connected provisioned device.

        Note: Connection to the target will be closed.
        :raises SPSDKTpTargetError: Cannot reset the target.
        """
        try:
            if not self.mboot.reset(reopen=False):
                raise SPSDKTpTargetError("Cannot reset connected target.")
        except (SPSDKTpTargetError, ValueError, McuBootError) as exc:
            raise SPSDKTpTargetError(
                f"Cannot reset connected target. Error code: {self.mboot.status_string}"
            ) from exc

    def load_sb_file(self, sb_file: bytes, timeout: Optional[int] = None) -> None:
        """Load SB file into provisioned device.

        :param sb_file: SB file data to be loaded into provisioned device.
        :param timeout: Timeout of operation in milliseconds.
        :raises SPSDKTpTargetError: Problem with loading SB image into target.
        """
        if not self.mboot.receive_sb_file(sb_file):
            raise SPSDKTpTargetError(
                f"The loading of SB file to target failed. Error code: {self.mboot.status_string}"
            )

    def prove_genuinity_challenge(self, challenge: bytes, timeout: Optional[int] = None) -> bytes:
        """Prove genuinity and get back the TP response to continue process of TP.

        :param challenge: Challenge data to start TP process.
        :param timeout: Timeout of operation in milliseconds.
        :return: Trust provisioning response for TP process.
        :raises SPSDKTpTargetError: Problem with genuinity challenge.
        """
        if not self.mboot.write_memory(self.buffer_address, challenge):
            raise SPSDKTpTargetError(
                f"Setting of challenge failed. Error code: {self.mboot.status_string}"
            )

        tp_response_length = self.mboot.tp_prove_genuinity(self.buffer_address, self.buffer_size)
        if tp_response_length is None:
            raise SPSDKTpTargetError(
                f"Executing Prove Genuinity failed. Error code: {self.mboot.status_string}"
            )

        ret = self.mboot.read_memory(self.buffer_address, tp_response_length)
        if not ret:
            raise SPSDKTpTargetError(
                f"Reading of Trusted provisioning failed. Error code: {self.mboot.status_string}"
            )

        return ret

    def set_wrapped_data(self, wrapped_data: bytes, timeout: Optional[int] = None) -> None:
        """Provide wrapped data to provisioned device.

        :param wrapped_data: Wrapped data to finish TP process.
        :param timeout: Timeout of operation in milliseconds.
        :raises SPSDKTpTargetError: Problem with memory writing.
        """
        if not self.mboot.write_memory(self.buffer_address, wrapped_data):
            raise SPSDKTpTargetError(
                f"Setting of wrapped data failed. Error code: {self.mboot.status_string}"
            )

        if not self.mboot.tp_set_wrapped_data(self.buffer_address):
            raise SPSDKTpTargetError(
                f"Executing the setting of OEM data failed. Error code: {self.mboot.status_string}"
            )

    @staticmethod
    def get_help() -> str:
        """Return help for this interface, including settings description."""
        return "\n".join(
            [
                "The BLHOST adapter settings allow better specified the BLHOST target, otherwise the",
                "default settings are used - This could lead to find a multiple BLHOST targets connected to system.",
                "The BLHOST target adapter settings:",
                "   - blhost_usb - BLHOST USB device (vid:pid).",
                "   - blhost_port - BLHOST UART device port.",
                "   - blhost_baudrate - BLHOST UART device port baudrate.",
                "   - blhost_timeout - BLHOST device atomic operations timeout.",
                "   - buffer_address - Address in memory used for communication (buffer exchange)",
                "   - buffer_size - Size of the communication buffer; 1kB by default)",
            ]
        )

    @classmethod
    def get_validation_schemas(cls) -> List[Dict[str, Any]]:
        """Return all additional validation schemas for interface.

        return: List of all additional validation schemas.
        """
        sch_cfg_file = ValidationSchemas.get_schema_file(TP_SCH_FILE)

        return [sch_cfg_file["target_blhost"]]

    def read_memory(self, address: int, length: int, memory_id: int = 0) -> bytes:
        """Read data from target's memory.

        :param address: Start address
        :param length: Number of bytes to read
        :param memory_id: Memory ID, defaults to 0
        :raises SPSDKTpTargetError: In case of a MBoot failure
        :return: Data read from the target
        """
        data = self.mboot.read_memory(address=address, length=length, mem_id=memory_id)
        if not data:
            raise SPSDKTpTargetError(
                f"Unable to read memory (address=0x{address:08x}, length={length}) "
                f"Error: {self.mboot.status_string}"
            )
        return data

    def write_memory(self, address: int, data: bytes, memory_id: int = 0) -> None:
        """Write data to target's memory.

        :param address: Start address
        :param data: Data to write
        :param memory_id: Memory ID, defaults to 0
        :raises SPSDKTpTargetError: In case of a MBoot failure
        """
        if not self.mboot.write_memory(address=address, data=data, mem_id=memory_id):
            raise SPSDKTpTargetError(
                f"Unable to write data (address=0x{address:08x}) Error: {self.mboot.status_string}"
            )

    def erase_memory(self, address: int, length: int, memory_id: int = 0) -> None:
        """Erase target's memory.

        :param address: Start address
        :param length: Number of bytes to erase
        :param memory_id: Memory ID, defaults to 0
        :raises SPSDKTpTargetError: In case of a MBoot failure
        """
        if not self.mboot.flash_erase_region(address=address, length=length, mem_id=memory_id):
            raise SPSDKTpTargetError(
                f"Unable to erase memory (address=0x{address:08x}, length=0x{length:08x}"
            )

    def check_provisioning_firmware(self) -> bool:
        """Check whether the Provisioning Firmware booted properly.

        :raises SPSDKTpTargetError: In case of a MBoot failure
        :return: True if ProvFW booted up
        """
        try:
            # we expect this command to fail
            # TP_CONTAINERINVALID or FAIL errors mean ProvFW is running
            # UNKNOWN_COMMAND indicates ROM in running
            # any other error code should be re-raised
            self.prove_genuinity_challenge(bytes(self.mboot.DEFAULT_MAX_PACKET_SIZE))
        except SPSDKTpTargetError:
            if self.mboot.status_code in [StatusCode.TP_CONTAINERINVALID, StatusCode.FAIL]:
                return True
            if self.mboot.status_code == StatusCode.UNKNOWN_COMMAND:
                return False
            raise
        else:
            # this should never happen
            raise SPSDKTpTargetError("Check for ProvFW boot-up malfunctioned!")



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2021-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause
"""Main Debug Authentication Tool application."""

import contextlib
import logging
import os
import sys
from typing import Dict, Iterator, List

import click

from spsdk import SPSDK_DATA_FOLDER
from spsdk.apps.nxpdebugmbox import get_debug_probe_options_help
from spsdk.apps.utils import spsdk_logger
from spsdk.apps.utils.common_cli_options import CommandsTreeGroup, spsdk_apps_common_options
from spsdk.apps.utils.utils import catch_spsdk_error
from spsdk.debuggers.utils import DebugProbe, open_debug_probe
from spsdk.exceptions import SPSDKError
from spsdk.shadowregs import ShadowRegisters, enable_debug
from spsdk.utils.misc import load_configuration, write_file
from spsdk.utils.reg_config import RegConfig
from spsdk.utils.registers import Registers, RegsRegister

logger = logging.getLogger(__name__)

CONFIG_DIR = os.path.join(SPSDK_DATA_FOLDER, "shadowregs")
CONFIG_DATABASE = os.path.join(CONFIG_DIR, "database.yaml")


@contextlib.contextmanager
def _open_debug_probe(pass_obj: Dict) -> Iterator[DebugProbe]:
    """Method opens DebugProbe object based on input arguments.

    :param pass_obj: Input dictionary with arguments.
    :return: Active DebugProbe object.
    :raises SPSDKError: Raised with any kind of problems with debug probe.
    """
    interface = pass_obj["interface"]
    serial_no = pass_obj["serial_no"]
    debug_probe_params = pass_obj["debug_probe_params"]

    with open_debug_probe(
        interface=interface, serial_no=serial_no, debug_probe_params=debug_probe_params
    ) as probe:
        yield probe


@contextlib.contextmanager
def _open_shadow_registers(pass_obj: Dict, connect: bool = True) -> Iterator[ShadowRegisters]:
    """Method opens ShadowRegisters object based on input arguments.

    :param pass_obj: Input dictionary with arguments.
    :param connect: Create Shadow register object connected to device.
    :return: Active ShadowRegisters object.
    :raises SPSDKError: Raised with any kind of problems with debug probe.
    """
    device = pass_obj["device"]
    revision = pass_obj["revision"]

    if device not in RegConfig(CONFIG_DATABASE).devices.device_names:
        raise SPSDKError(
            "Invalid or none device parameter(-dev). Check '--help' to get supported devices."
        )

    regs_cfg = RegConfig(CONFIG_DATABASE)

    if not connect:
        yield ShadowRegisters(debug_probe=None, config=regs_cfg, device=device, revision=revision)
    else:
        with _open_debug_probe(pass_obj) as debug_probe:
            if not enable_debug(debug_probe):
                raise SPSDKError("Cannot enable debug interface")

            yield ShadowRegisters(
                debug_probe=debug_probe, config=regs_cfg, device=device, revision=revision
            )


@click.group(name="shadowregs", no_args_is_help=True, cls=CommandsTreeGroup)
@click.option(
    "-i",
    "--interface",
    help="The interface allow specify to use only one debug probe interface"
    " like: 'PyOCD', 'jlink' or 'pemicro'",
)
@spsdk_apps_common_options
@click.option(
    "-s",
    "--serial-no",
    help="Serial number of debug probe to avoid select menu after startup.",
)
@click.option(
    "-dev",
    "--device",
    type=click.Choice(RegConfig.get_devices(CONFIG_DATABASE).device_names),
    help="The target device family.",
)
@click.option(
    "-r",
    "--revision",
    help="Chip revision; if not specified, most recent one will be used",
)
@click.option(
    "-o",
    "--debug-probe-option",
    multiple=True,
    help=get_debug_probe_options_help(),
)
@click.pass_context
def main(
    ctx: click.Context,
    interface: str,
    log_level: int,
    serial_no: str,
    debug_probe_option: List[str],
    device: str,
    revision: str,
) -> int:
    """NXP Shadow Registers control Tool."""
    spsdk_logger.install(level=log_level)

    probe_user_params = {}
    for par in debug_probe_option:
        if par.count("=") != 1:
            raise SPSDKError(f"Invalid -o parameter {par}!")
        par_splitted = par.split("=")
        probe_user_params[par_splitted[0]] = par_splitted[1]

    ctx.obj = {
        "interface": interface,
        "serial_no": serial_no,
        "debug_probe_params": probe_user_params,
        "device": device,
        "revision": revision or "latest",
    }

    return 0


# Enable / Disable debug
@main.command()
@click.option(
    "-f",
    "--filename",
    type=click.Path(resolve_path=True),
    default="sr_config.yaml",
    help="The name of file used to save the current configuration."
    " Default name is 'sr_config'. The extension is always '.yaml'.",
)
@click.option(
    "-r",
    "--raw",
    is_flag=True,
    default=False,
    help="The stored configuration will include also the computed fields "
    "and anti-pole registers.",
)
@click.option(
    "-d",
    "--save-diff",
    is_flag=True,
    default=False,
    help="Save differences comparing to defaults",
)
@click.pass_obj
def saveconfig(pass_obj: dict, filename: str, raw: bool, save_diff: bool) -> None:
    """Save current state of shadow registers to YAML file."""
    try:
        with _open_shadow_registers(pass_obj) as shadow_regs:
            shadow_regs.reload_registers()
            shadow_regs.create_yaml_config(filename, raw, diff=save_diff)
        click.echo(f"The Shadow registers has been saved into {filename} YAML file")
    except SPSDKError as exc:
        raise SPSDKError(f"Save configuration of Shadow registers failed! ({str(exc)})") from exc


@main.command(no_args_is_help=True)
@click.option(
    "-f",
    "--filename",
    type=click.Path(resolve_path=True),
    required=True,
    help="The name of file used to load a new configuration."
    " Default name is 'sr_config'. The extension is always '.yaml'.",
)
@click.option(
    "-r",
    "--raw",
    is_flag=True,
    default=False,
    help="In loaded configuration will accepted also the computed fields "
    "and anti-pole registers.",
)
@click.option(
    "--verify/--no-verify",
    is_flag=True,
    default=True,
    help="Verify write operation (verify by default)",
)
@click.pass_obj
def loadconfig(pass_obj: dict, filename: str, raw: bool, verify: bool) -> None:
    """Load new state of shadow registers from YAML file into microcontroller."""
    try:
        with _open_shadow_registers(pass_obj) as shadow_regs:
            shadow_regs.load_yaml_config(filename, raw)
            shadow_regs.sets_all_registers(verify)
        click.echo(f"The Shadow registers has been loaded by configuration in {filename} YAML file")
    except SPSDKError as exc:
        raise SPSDKError(f"Load configuration of Shadow registers failed ({str(exc)})!") from exc


@main.command(name="get-template", no_args_is_help=True)
@click.option(
    "-r",
    "--raw",
    is_flag=True,
    default=False,
    help="In loaded configuration will accepted also the computed fields "
    "and anti-pole registers.",
)
@click.argument("output", metavar="PATH", type=click.Path(resolve_path=True))
@click.pass_obj
def get_template(pass_obj: dict, output: str, raw: bool) -> None:
    """Generate the template of Shadow registers YAML configuration file.

    \b
    PATH    - file name path to write template config file
    """
    with _open_shadow_registers(pass_obj, connect=False) as shadow_regs:
        shadow_regs.create_yaml_config(output, raw)
    click.echo(
        f"The Shadow registers template for {pass_obj['device']} has been saved into {output} YAML file"
    )


@main.command()
@click.option(
    "-r",
    "--rich",
    is_flag=True,
    default=False,
    help="Enables rich format of printed output.",
)
@click.pass_obj
def printregs(pass_obj: dict, rich: bool = False) -> None:
    """Print all Shadow registers including theirs current values.

    In case of needed more information, there is also provided rich format of print.
    """
    try:
        with _open_shadow_registers(pass_obj) as shadow_regs:
            shadow_regs.reload_registers()

            for reg in shadow_regs.regs.get_registers():
                click.echo(f"Register Name:        {reg.name}")
                click.echo(f"Register value:       {reg.get_hex_value()}")
                if rich:
                    click.echo(f"Register description: {reg.description}")
                    address = shadow_regs.offset + reg.offset
                    click.echo(f"Register address:     0x{address:08X}")
                    click.echo(f"Register width:       {reg.width} bits")
                click.echo()
    except SPSDKError as exc:
        raise SPSDKError(f"Print of Shadow registers failed! ({str(exc)})") from exc


@main.command(no_args_is_help=True)
@click.option("-r", "--reg", type=str, required=True, help="The name of register to be read.")
@click.pass_obj
def getreg(pass_obj: dict, reg: str) -> None:
    """The command prints the current value of one shadow register."""
    try:
        with _open_shadow_registers(pass_obj) as shadow_regs:
            register: RegsRegister = shadow_regs.regs.find_reg(reg, include_group_regs=True)
            value = shadow_regs.read_register(register)
            click.echo(f"Value of {reg} is: {value.hex()}")
    except SPSDKError as exc:
        raise SPSDKError(f"Getting Shadow register failed! ({str(exc)})") from exc


@main.command(no_args_is_help=True)
@click.option("-r", "--reg", type=str, required=True, help="The name of register to be set.")
@click.option(
    "-v", "--reg-val", type=str, required=True, help="The new value of register in hex format."
)
@click.option(
    "--verify/--no-verify",
    is_flag=True,
    default=True,
    help="Verify write operation (verify by default)",
)
@click.pass_obj
def setreg(pass_obj: dict, reg: str, reg_val: str, verify: bool) -> None:
    """The command sets a value of one shadow register defined by parameter."""
    try:
        with _open_shadow_registers(pass_obj) as shadow_regs:
            shadow_regs.set_register(reg, reg_val, verify)
        click.echo(f"The Shadow register {reg} has been set to {reg_val} value")
    except SPSDKError as exc:
        raise SPSDKError(f"Setting Shadow register failed! ({str(exc)})") from exc


@main.command()
@click.pass_obj
def reset(pass_obj: dict) -> None:
    """The command resets connected device."""
    with _open_debug_probe(pass_obj) as probe:
        probe.reset()
    click.echo("The target has been reset.")


@main.command()
@click.option(
    "-o",
    "--output",
    type=click.Path(resolve_path=True),
    required=True,
    help="File name of generated output HTML description file",
)
@click.option(
    "-p",
    "--open",
    "open_result",
    is_flag=True,
    help="Open the generated description file",
)
@click.pass_obj
def info(pass_obj: dict, output: str, open_result: bool) -> None:
    """The command generate HTML of Shadow registers."""
    config = RegConfig(CONFIG_DATABASE)
    device = pass_obj["device"]
    revision = pass_obj["revision"]
    registers = Registers(device)
    rev = (
        revision
        if revision != "latest"
        else config.devices.get_by_name(device).revisions.get_latest().name
    )
    registers.load_registers_from_xml(config.get_data_file(device, rev))
    html_output = registers.generate_html(
        f"{device} - Shadow Registers",
        f"The table with Shadow registers description for {device}",
    )
    write_file(html_output, output, encoding="utf-8")
    if open_result:  # pragma: no cover # can't test opening the html document
        click.launch(f"{output}")


@main.command(name="fuses-script", no_args_is_help=True)
@click.option(
    "-c",
    "--config",
    type=click.Path(),
    help="The name of shadow register configuration file used to generate burn fuses BLHOST script.",
    required=True,
)
@click.argument(
    "output",
    type=click.Path(resolve_path=True),
    required=True,
)
@click.pass_obj
def fuses_script(pass_obj: dict, config: str, output: str) -> None:
    """The command generate BLHOST script to burn up fuses in device by configuration.

    The OUTPUT argument specify file name of generate BLHOST burn fuses script.
    """
    try:
        with _open_shadow_registers(pass_obj, connect=False) as shadow_regs:
            shadow_regs.load_yaml_config(config, False)
            cfg = load_configuration(config)
            write_file(shadow_regs.create_fuse_blhost_script(list(cfg["registers"].keys())), output)
        click.echo(f"BLHOST script to burn fuses has been generated: {output}")
    except SPSDKError as exc:
        raise SPSDKError(
            f"Creating BLHOST script from shadow register configuration failed! ({str(exc)})"
        ) from exc


@catch_spsdk_error
def safe_main() -> None:
    """Safe main method."""
    sys.exit(main())  # pylint: disable=no-value-for-parameter


if __name__ == "__main__":
    safe_main()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2020-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Module provides support for checking the brick-conditions in PFR settings."""

import logging
import os
from dataclasses import dataclass
from typing import List, Optional, Tuple

from spsdk.pfr import PFR_DATA_FOLDER, Processor, Translator
from spsdk.pfr.exceptions import SPSDKPfrcMissingConfigError, SPSDKPfrConfigError
from spsdk.utils.database import Database
from spsdk.utils.misc import load_configuration

from .pfr import CFPA, CMPA, SPSDKPfrError

PFRC_DATA_FOLDER = os.path.join(PFR_DATA_FOLDER, "pfrc")
PFRC_DATABASE_FILE = os.path.join(PFRC_DATA_FOLDER, "database.yaml")

logger = logging.getLogger(__name__)


@dataclass
class Rule:
    """Dataclass holding information about individual rule."""

    req_id: str
    desc: str
    msg: str
    cond: str


class RulesList(List[Rule]):
    """List of rules."""

    @staticmethod
    def load_from_files_list(rules_file_list: list) -> "RulesList":
        """Load the rules from list of files.

        :param rules_file_list: A list of paths to configuration files containing rules to be loaded
        :returns RulesList object with loaded rules
        """
        rules_list = RulesList()
        for rules_file in rules_file_list:
            rules = RulesList.load_from_file(rules_file)
            rules_list.extend(rules)
        return rules_list

    @staticmethod
    def load_from_file(rules_file: str) -> "RulesList":
        """Load the rules from a single file.

        :param rules_file: A path to a configuration file containing rules to be loaded
        :returns RulesList object with loaded rules
        """
        rules = load_configuration(rules_file)
        return RulesList(Rule(**rule) for rule in rules)


class Pfrc:
    """Class responsible for checking of the conditions."""

    def __init__(
        self,
        cmpa: Optional[CMPA] = None,
        cfpa: Optional[CFPA] = None,
    ) -> None:
        """Initialize an instance.

        :param cmpa: configuration data loaded from CMPA config file, defaults to None
        :param cfpa: configuration data loaded from CFPA config file, defaults to None
        :raises SPSDKPfrError: No configuration is provided
        :raises SPSDKPfrConfigError: Problem with PFR configuration(s) occured
        """
        self.database = Database(PFRC_DATABASE_FILE)
        if not (cmpa or cfpa):
            raise SPSDKPfrError("No cmpa or cfpa configurations specified")
        self.cmpa = cmpa
        self.cfpa = cfpa
        self.chip_family = cmpa.device if cmpa else cfpa.device  # type: ignore
        if not self.chip_family or self.chip_family not in self.get_supported_families():
            raise SPSDKPfrConfigError(
                f"Chip family from configuration is not supported: {self.chip_family} "
                f"Supported families:{self.get_supported_families()}"
            )

    @staticmethod
    def get_supported_families() -> List[str]:
        """Return list of supported families.

        :return: List of supported families.
        """
        database = Database(PFRC_DATABASE_FILE)
        return database.devices.device_names

    def validate_brick_conditions(
        self, additional_rules_file: Optional[str] = None
    ) -> Tuple[RulesList, RulesList, RulesList]:
        """The method validates the brick conditions for specified configuration.

        :param additional_rules_file: Additional rules file, defaults to None
        :returns Tuple with passed, failed and skipped rules as a RulesList
        :raises SPSDKPfrError: Brick condition validation failed
        """
        rules = self.load_rules(additional_rules_file)
        translator = Translator(cmpa=self.cmpa, cfpa=self.cfpa)
        processor = Processor(translator=translator)
        try:
            failed_rules, passed_rules, skipped_rules = RulesList(), RulesList(), RulesList()
            for rule in rules:
                logger.debug(f"Processing the brick condition: {rule.req_id}: {rule.desc} ")
                try:
                    result, condition = processor.process(rule.cond)
                except SPSDKPfrcMissingConfigError as e:
                    skipped_rules.append(rule)
                    logger.debug(f"Brick condition not translated: {e}")
                    continue
                logger.debug(f"Brick condition '{rule.cond}' translated to '{condition}'")
                if result:
                    failed_rules.append(rule)
                    logger.warning(f"FAIL: you are going to brick your device\n{rule.msg}")
                else:
                    passed_rules.append(rule)
                    logger.info("OK: Brick condition not fulfilled")
        except SyntaxError as e:
            raise SPSDKPfrError(f"\nERROR: Unable to parse: '{e}'") from e
        except (KeyError, ValueError, TypeError) as e:
            raise SPSDKPfrError(f"\nERROR: Unable to lookup identifier: {e}") from e
        except Exception as e:  # pylint: disable=broad-except
            raise SPSDKPfrError(f"Error e({e}) while evaluating {rule.cond}") from e
        return passed_rules, failed_rules, skipped_rules

    def load_rules(self, additional_rules_file: Optional[str] = None) -> RulesList:
        """The function loads the rules for family and optionally add additional rules from user.

        :param additional_rules_file: Additional rules file, defaults to None
        :return: Loaded rules in list of dictionaries.
        """
        rules_files = self.database.get_device_value("rules", self.chip_family)
        rules_files = [os.path.join(PFRC_DATA_FOLDER, rules_file) for rules_file in rules_files]
        if additional_rules_file:
            rules_files.append(additional_rules_file)
        return RulesList.load_from_files_list(rules_files)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2020-2022 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Secure Boot Header."""

from datetime import datetime
from struct import calcsize, pack, unpack_from
from typing import Optional

from spsdk import SPSDKError
from spsdk.sbfile.misc import pack_timestamp, unpack_timestamp
from spsdk.utils.crypto.abstract import BaseClass
from spsdk.utils.crypto.common import crypto_backend
from spsdk.utils.easy_enum import Enum
from spsdk.utils.misc import DebugInfo, swap16

from ..misc import BcdVersion3, BcdVersion3Format, SecBootBlckSize
from ..sb2.commands import CmdHeader, CmdTag


class SecureBootFlagsV1(Enum):
    """Flags for SectionHeader."""

    NONE = (0, "No flags")
    ROM_SECTION_BOOTABLE = (
        1,
        "The section is bootable and contains a sequence of bootloader commands.",
    )
    ROM_SECTION_CLEARTEXT = (
        2,
        "The section is unencrypted. Applies only if the rest of the boot image is encrypted.",
    )


########################################################################################################################
# Secure Boot Header Class (Version SB1.x)
########################################################################################################################
# pylint: disable=too-many-instance-attributes
class SecureBootHeaderV1(BaseClass):
    """Secure Boot Header V1."""

    # binary format of the header
    _FORMAT = "<20s4s2BH3I5H2s4sQ13H6s"
    # size of the header in bytes
    _SIZE = calcsize(_FORMAT)
    # binary signature 1 for header identification
    _SIGNATURE1 = b"STMP"
    # binary signature 2 for header identification
    _SIGNATURE2 = b"sgtl"

    def __init__(
        self,
        version: str = "1.0",
        product_version: BcdVersion3Format = BcdVersion3.DEFAULT,
        component_version: BcdVersion3Format = BcdVersion3.DEFAULT,
        flags: int = 0,
        drive_tag: int = 0,
        digest: bytes = b"\0" * 20,
        timestamp: Optional[datetime] = None,
    ) -> None:
        """Initialize SecureBootHeaderV1.

        :param version: of the format: 1.0 or 1.1 or 1.2
        :param product_version: Product version.
        :param component_version: Component version.
        :param flags: Flags associated with the entire image.
        :param drive_tag: Identifier for the disk drive or partition containing this image.
        :param digest: SHA-1 digest of all fields of the header, 20 bytes
                    The first 16 bytes (of 20 total) also act as the initialization vector for CBC-encrypted regions.
        :param timestamp: datetime of the file creation, use None for current date/time
                    Fixed value should be used only for regression testing to generate same results
        :raises SPSDKError: Invalid header version
        """
        # SHA-1 digest of all fields of the header prior to this one.
        self.digest = digest
        # Major version of the boot image format, currently 1. Minor version of the boot image format, currently 1 or 2.
        if version not in ("1.0", "1.1", "1.2"):
            raise SPSDKError("Invalid header version")
        self.version = version
        self.flags = flags
        # Size of the entire image in blocks.
        self.image_blocks = 0
        # Unique identifier of the section to start booting from.
        self.first_boot_section_id = 0
        # Number of entries in the DEK dictionary.
        self.key_count = 0
        # Size of the entire image header in blocks. This seems to be a constant.
        self.header_blocks = SecBootBlckSize.to_num_blocks(self._SIZE)
        # Number of sections.
        self.section_count = 0
        # Size in blocks of a section header. This seems to be a constant.
        self.section_header_size = SecBootBlckSize.to_num_blocks(BootSectionHeaderV1.SIZE)
        # Timestamp in microseconds size 1-1-2000 00:00 when the image was created.
        self.timestamp = (
            timestamp if timestamp else datetime.fromtimestamp(int(datetime.now().timestamp()))
        )
        # Product version in format #.#.#
        self.product_version: BcdVersion3 = BcdVersion3.to_version(product_version)
        # Component version in format #.#.#
        self.component_version: BcdVersion3 = BcdVersion3.to_version(component_version)
        # Identifier for the disk drive or partition containing this image.
        self.drive_tag = drive_tag

    def __str__(self) -> str:
        return f"Header: v{self.version}, {self.image_blocks}"

    @property
    def key_dictionary_block(self) -> int:
        """Return key dictionary block."""
        return self.header_blocks + self.section_count * self.section_header_size

    @property
    def first_boot_tag_block(self) -> int:
        """Return first boot tag block."""
        return self.key_dictionary_block + self.key_count * 2

    @property
    def size(self) -> int:
        """Return size of the header in bytes."""
        return self._SIZE

    def info(self) -> str:
        """Get info of Header as a string."""
        return (
            f" Digest:               {self.digest.hex()}\n"
            f" Version:              {self.version}\n"
            f" Flags:                0x{self.flags:04X}\n"
            f" Image Blocks:         {self.image_blocks}\n"
            f" First Boot Tag Block: {self.first_boot_tag_block}\n"
            f" First Boot SectionID: {self.first_boot_section_id}\n"
            f" Key Count:            {self.key_count}\n"
            f" Key Dictionary Block: {self.key_dictionary_block}\n"
            f" Header Blocks:        {self.header_blocks}\n"
            f" Section Count:        {self.section_count}\n"
            f" Section Header Size:  {self.section_header_size}\n"
            f" Timestamp:            {self.timestamp}\n"
            f" Product Version:      {self.product_version}\n"
            f" Component Version:    {self.component_version}\n"
            f" Drive Tag:            {self.drive_tag}\n"
        )

    def export(
        self,
        padding8: Optional[bytes] = None,
        dbg_info: DebugInfo = DebugInfo.disabled(),
    ) -> bytes:
        """Serialization to binary form.

        :param padding8: 8 padding bytes used for in the header, None to use random bytes
                This value shall be used only for regression testing to generate same results
        :param dbg_info: class allowing to debug output from the export
        :return: Serialize object into bytes
        """
        major_version, minor_version = [int(v) for v in self.version.split(".")]
        product_version_words = [swap16(n) for n in self.product_version.nums]
        component_version_words = [swap16(n) for n in self.component_version.nums]
        signature2 = crypto_backend().random_bytes(4)
        padding = padding8 if padding8 else crypto_backend().random_bytes(8)

        if (major_version > 1) or ((major_version == 1) and (minor_version >= 2)):
            signature2 = self._SIGNATURE2

        dbg_info.append_section("SB-file-Header")

        result = pack(
            self._FORMAT,
            self.digest,
            self._SIGNATURE1,
            # header version
            major_version,
            minor_version,
            self.flags,
            self.image_blocks,
            self.first_boot_tag_block,
            self.first_boot_section_id,
            self.key_count,
            self.key_dictionary_block,
            self.header_blocks,
            self.section_count,
            self.section_header_size,
            padding[0:2],
            signature2,
            pack_timestamp(self.timestamp),
            # product version
            product_version_words[0],
            0,
            product_version_words[1],
            0,
            product_version_words[2],
            0,
            # component version
            component_version_words[0],
            0,
            component_version_words[1],
            0,
            component_version_words[2],
            0,
            self.drive_tag,
            padding[2:],
        )

        result = result[len(self.digest) :]
        self.digest = crypto_backend().hash(result, "sha1")

        dbg_info.append_binary_section("digest", self.digest)
        dbg_info.append_binary_section("attrs", result)

        return self.digest + result

    # pylint: disable=too-many-locals
    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> "SecureBootHeaderV1":
        """Convert binary data into the instance (deserialization).

        :param data: given binary data to be decoded
        :param offset: to start parsing binary data; 0 by default
        :return: the instance of secure boot header v1
        :raises SPSDKError: Raised when there is insufficient size
        :raises SPSDKError: Raised when there is invalid signature
        :raises SPSDKError: Raised when there is unexpected signature
        """
        if SecureBootHeaderV1._SIZE > len(data) - offset:
            raise SPSDKError("Insufficient size")

        (
            digest,
            signature1,
            # header version
            major_version,
            minor_version,
            flags,
            image_blocks,
            _first_boot_tag_block,
            first_boot_section_id,
            key_count,
            _key_dictionary_block,
            header_blocks,
            section_count,
            section_header_size,
            _,  # padding 2
            signature2,
            timestamp,
            pv0,
            _,
            pv1,
            _,
            pv2,
            _,  # product version
            cv0,
            _,
            cv1,
            _,
            cv2,
            _,  # component version
            drive_tag,
            _,  # padding 6
        ) = unpack_from(SecureBootHeaderV1._FORMAT, data, offset)

        # check header signature 1
        if signature1 != SecureBootHeaderV1._SIGNATURE1:
            raise SPSDKError("Invalid signature")

        # check header signature 2 for version 1.1 and greater
        if (major_version > 1) or ((major_version == 1) and (minor_version >= 2)):
            if signature2 != SecureBootHeaderV1._SIGNATURE2:
                raise SPSDKError("Unexpected signature")

        product_version = BcdVersion3(swap16(pv0), swap16(pv1), swap16(pv2))
        component_version = BcdVersion3(swap16(cv0), swap16(cv1), swap16(cv2))

        obj = SecureBootHeaderV1(
            digest=digest,
            version=f"{major_version}.{minor_version}",
            flags=flags,
            product_version=product_version,
            component_version=component_version,
            drive_tag=drive_tag,
        )

        obj.image_blocks = image_blocks
        obj.first_boot_section_id = first_boot_section_id
        obj.key_count = key_count  # key_blob_block = key_count
        obj.header_blocks = header_blocks
        obj.section_count = section_count
        obj.section_header_size = section_header_size
        obj.timestamp = unpack_timestamp(timestamp)

        return obj


class SectionHeaderItemV1(BaseClass):
    """Section header item/row in section header table in in SB file V1.x."""

    FORMAT = "<4I"
    SIZE = calcsize(FORMAT)

    def __init__(
        self,
        identifier: int = 0,
        offset: int = 0,
        num_blocks: int = 0,
        flags: SecureBootFlagsV1 = SecureBootFlagsV1.NONE,
    ):
        """Initialize SectionHeaderItemV1.

        :param identifier: Unique 32-bit identifier for this section.
        :param offset: The starting cipher block for this section's data from the beginning of the image.
        :param num_blocks: The length of the section data in cipher blocks.
        :param flags: Flags that apply to the entire section, see SectionHeaderV1Flags
        """
        self.identifier = identifier
        self.offset = offset
        self.num_blocks = num_blocks
        self._flags = flags

    def __str__(self) -> str:
        return (
            f"SectionHeaderV1: ID={self.identifier}, Ofs={self.offset}, NumBlocks={self.num_blocks}, "
            f"Flag=0x{self._flags:X}"
        )

    @property
    def flags(self) -> int:
        """Return flags, see SectionHeaderV1Flags."""
        return self._flags

    @property
    def bootable(self) -> bool:
        """Return whether section is bootable."""
        return self._flags & SecureBootFlagsV1.ROM_SECTION_BOOTABLE != 0

    @property
    def size(self) -> int:
        """Return size of exported data in bytes."""
        return self.SIZE

    def info(self) -> str:
        """Return Get text info of Header."""
        return (
            f" Identifier: 0x{self.identifier:08X}\n"
            f" Offset:     {self.offset}\n"
            f" NumBlocks:  {self.num_blocks}\n"
            f" Bootable:   {'YES' if self.bootable else 'NO'}\n"
        )

    def export(self) -> bytes:
        """Return serialization to binary format."""
        return pack(self.FORMAT, self.identifier, self.offset, self.num_blocks, self._flags)

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> "SectionHeaderItemV1":
        """Parse binary data into the instance (deserialization).

        :param data: to be parsed
        :param offset: to start parsing the data
        :return: the new instance
        :raises SPSDKError: If size is not sufficient
        """
        if cls.SIZE > len(data) - offset:
            raise SPSDKError("Insufficient size")
        (identifier, offset, length, flags) = unpack_from(cls.FORMAT, data, offset)
        return cls(identifier, offset, length, SecureBootFlagsV1.from_int(flags))


class BootSectionHeaderV1(CmdTag):
    """Header of boot section derived from command TAG.

    Note: Command TAG was reused to save some code in Boot ROM
    """

    # size of the binary representation of the header in bytes
    SIZE = CmdHeader.SIZE

    # Mask for header flag, that identifies ROM_LAST_TAG flag
    _ROM_LAST_TAG_MASK = 1

    def __init__(self, section_id: int = 0, flags: SecureBootFlagsV1 = SecureBootFlagsV1.NONE):
        """Initialize BootSectionHeaderV1.

        :param section_id: unique section ID, 32-bit int
        :param flags: see SecureBootFlagsV1
        """
        super().__init__()
        self.header.address = section_id
        self.header.flags = 0
        self.header.data = (
            flags  # not sure here, it seems flags are duplicates as 32-bit integer too???
        )

    @property
    def section_id(self) -> int:
        """Return unique ID of the section, 32 number."""
        return self.header.address

    @property
    def num_blocks(self) -> int:
        """Return size of the section in number of cipher blocks."""
        return self.header.count

    @num_blocks.setter
    def num_blocks(self, value: int) -> None:
        """Setter.

        :param value: size of the section in number of cipher blocks
        """
        self.header.count = value

    @property
    def rom_last_tag(self) -> bool:
        """Return ROM_LAST_TAG flag.

        The last section header in an image always has its ROM_LAST_TAG flag set to help the ROM know at what point
        to stop searching.
        """
        return self.header.flags & self._ROM_LAST_TAG_MASK != 0

    @rom_last_tag.setter
    def rom_last_tag(self, value: bool) -> None:
        """Setter.

        :param value: ROM_LAST_TAG flag
        """
        if value:
            self.header.flags |= self._ROM_LAST_TAG_MASK
        else:
            self.header.flags &= ~self._ROM_LAST_TAG_MASK

    @property
    def flags(self) -> SecureBootFlagsV1:
        """Return section flags."""
        return SecureBootFlagsV1.from_int(self.header.data)

    @property
    def bootable(self) -> bool:
        """Return whether section is bootable."""
        return self.flags & SecureBootFlagsV1.ROM_SECTION_BOOTABLE != 0

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> "BootSectionHeaderV1":
        """Parse from bytes into BootSectionHeaderV1 object."""
        cmd_tag = super(BootSectionHeaderV1, cls).parse(data, offset)
        assert isinstance(cmd_tag, BootSectionHeaderV1)
        return cmd_tag



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2022-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause
"""This module contains generic implementation of image segment."""
import abc
from typing import Any, Dict

from spsdk.exceptions import SPSDKValueError
from spsdk.utils.database import Database
from spsdk.utils.registers import Registers


class SegmentBase:
    """Base class for image segment."""

    def __init__(self, family: str, revision: str) -> None:
        """Segment base Constructor.

        :param family: Chip family.
        :param revision: Optional Chip family revision.
        :raises SPSDKValueError: Unsupported family.
        """
        if family not in self.get_supported_families():
            raise SPSDKValueError(
                f"Unsupported chip family:{family}. {family} not in {self.get_supported_families()}"
            )
        self.family = family
        self.revision = revision

    @property
    @abc.abstractmethod
    def registers(self) -> Registers:
        """Registers of segment."""

    def export(self) -> bytes:
        """Export block binary.

        :return: Binary representation of segment.
        """
        return self.registers.image_info().export()

    def parse(self, binary: bytes) -> None:
        """Parse binary block into segment object.

        :param binary: binary image.
        :raises SPSDKValueError: Invalid input binary length.
        """
        if len(binary) != len(self.registers.image_info()):
            raise SPSDKValueError(
                f"Invalid length of input binary: {len(binary)} != {len(self.registers.image_info())}"
            )
        for reg in self.registers.get_registers():
            reg.set_value(
                int.from_bytes(binary[reg.offset // 8 : reg.offset // 8 + reg.width // 8], "little")
            )

    @staticmethod
    @abc.abstractmethod
    def load_from_config(config: Dict) -> Any:
        """Load configuration file.

        :param config: XMCD configuration file.
        :return: Segment object.
        """

    @abc.abstractmethod
    def create_config(self) -> str:
        """Create current configuration YAML.

        :return: Configuration of segment.
        """

    @classmethod
    def get_supported_families(cls) -> list:
        """Return list of supported families.

        :return: List of supported families.
        """
        return cls.get_database().devices.device_names

    @classmethod
    def get_memory_types(cls, family: str, revision: str = "latest") -> dict:
        """Get memory types data from database.

        :param family: Chip family.
        :param revision: Optional Chip family revision.
        """
        return cls.get_database().get_device_value("mem_types", family, revision, default={})

    @classmethod
    def get_supported_memory_types(cls, family: str, revision: str = "latest") -> list:
        """Get list of supported memory types data from database.

        :param family: Chip family.
        :param revision: Optional Chip family revision.
        """
        return list(cls.get_memory_types(family, revision).keys())

    @staticmethod
    @abc.abstractmethod
    def get_database() -> Database:
        """Get the devices database."""
